"""Wrapper functions for use in the config API.

All of the functions in this module can be listed in the config passed to
:code:`osl_dynamics.run_pipeline`.

All wrapper functions have the structure::

    func(data, output_dir, **kwargs)

where:

- :code:`data` is an :code:`osl_dynamics.data.Data` object.
- :code:`output_dir` is the path to save output to.
- :code:`kwargs` are keyword arguments for function specific options.
"""

import os
import logging
from pathlib import Path

import numpy as np
import matplotlib.pyplot as plt

from osl_dynamics import array_ops
from osl_dynamics.utils.misc import load, override_dict_defaults, save

_logger = logging.getLogger("osl-dynamics")


def load_data(inputs, kwargs=None, prepare=None):
    """Load and prepare data.

    Parameters
    ----------
    inputs : str
        Path to directory containing :code:`npy` files.
    kwargs : dict, optional
        Keyword arguments to pass to the `Data <https://osl-dynamics\
        .readthedocs.io/en/latest/autoapi/osl_dynamics/data/index.html\
        #osl_dynamics.data.Data>`_ class. Useful keyword arguments to pass are
        :code:`sampling_frequency`, :code:`mask_file` and
        :code:`parcellation_file`.
    prepare : dict, optional
        Methods dict to pass to the prepare method. See docstring for
        `Data <https://osl-dynamics.readthedocs.io/en/latest/autoapi\
        /osl_dynamics/data/index.html#osl_dynamics.data.Data>`_.prepare.

    Returns
    -------
    data : osl_dynamics.data.Data
        Data object.
    """
    from osl_dynamics.data import Data

    kwargs = {} if kwargs is None else kwargs
    prepare = {} if prepare is None else prepare

    data = Data(inputs, **kwargs)
    data.prepare(prepare)
    return data


def train_hmm(
    data,
    output_dir,
    config_kwargs,
    init_kwargs=None,
    fit_kwargs=None,
    save_inf_params=True,
):
    """Train a `Hidden Markov Model <https://osl-dynamics.readthedocs.io/en\
    /latest/autoapi/osl_dynamics/models/hmm/index.html>`_.

    This function will:

    1. Build an :code:`hmm.Model` object.
    2. Initialize the parameters of the model using
       :code:`Model.random_state_time_course_initialization`.
    3. Perform full training.
    4. Save the inferred parameters (state probabilities, means and covariances)
       if :code:`save_inf_params=True`.

    This function will create two directories:

    - :code:`<output_dir>/model`, which contains the trained model.
    - :code:`<output_dir>/inf_params`, which contains the inferred parameters.
      This directory is only created if :code:`save_inf_params=True`.

    Parameters
    ----------
    data : osl_dynamics.data.Data
        Data object for training the model.
    output_dir : str
        Path to output directory.
    config_kwargs : dict
        Keyword arguments to pass to `hmm.Config <https://osl-dynamics\
        .readthedocs.io/en/latest/autoapi/osl_dynamics/models/hmm/index.html\
        #osl_dynamics.models.hmm.Config>`_. Defaults to::

            {'sequence_length': 200,
             'batch_size': 256,
             'learning_rate': 0.01,
             'n_epochs': 20}.
    init_kwargs : dict, optional
        Keyword arguments to pass to
        :code:`Model.random_state_time_course_initialization`. Defaults to::

            {'n_init': 5, 'n_epochs': 2}.
    fit_kwargs : dict, optional
        Keyword arguments to pass to the :code:`Model.fit`. No defaults.
    save_inf_params : bool, optional
        Should we save the inferred parameters?
    """
    if data is None:
        raise ValueError("data must be passed.")

    from osl_dynamics.models import hmm

    init_kwargs = {} if init_kwargs is None else init_kwargs
    fit_kwargs = {} if fit_kwargs is None else fit_kwargs

    # Directories
    model_dir = output_dir + "/model"

    # Create the model object
    _logger.info("Building model")
    default_config_kwargs = {
        "n_channels": data.n_channels,
        "sequence_length": 200,
        "batch_size": 256,
        "learning_rate": 0.01,
        "n_epochs": 20,
    }
    config_kwargs = override_dict_defaults(default_config_kwargs, config_kwargs)
    _logger.info(f"Using config_kwargs: {config_kwargs}")
    config = hmm.Config(**config_kwargs)
    model = hmm.Model(config)
    model.summary()

    # Initialisation
    default_init_kwargs = {"n_init": 5, "n_epochs": 2}
    init_kwargs = override_dict_defaults(default_init_kwargs, init_kwargs)
    _logger.info(f"Using init_kwargs: {init_kwargs}")
    init_history = model.random_state_time_course_initialization(
        data,
        **init_kwargs,
    )

    # Training
    history = model.fit(data, **fit_kwargs)

    # Get the variational free energy
    history["free_energy"] = model.free_energy(data)

    # Save trained model
    _logger.info(f"Saving model to: {model_dir}")
    model.save(model_dir)
    save(f"{model_dir}/init_history.pkl", init_history)
    save(f"{model_dir}/history.pkl", history)

    if save_inf_params:
        # Make output directory
        inf_params_dir = output_dir + "/inf_params"
        os.makedirs(inf_params_dir, exist_ok=True)

        # Get the inferred parameters
        alpha = model.get_alpha(data)
        means, covs = model.get_means_covariances()

        # Save inferred parameters
        save(f"{inf_params_dir}/alp.pkl", alpha)
        save(f"{inf_params_dir}/means.npy", means)
        save(f"{inf_params_dir}/covs.npy", covs)


def train_dynemo(
    data,
    output_dir,
    config_kwargs,
    init_kwargs=None,
    fit_kwargs=None,
    save_inf_params=True,
):
    """Train `DyNeMo <https://osl-dynamics.readthedocs.io/en/latest/autoapi\
    /osl_dynamics/models/dynemo/index.html>`_.

    This function will:

    1. Build a :code:`dynemo.Model` object.
    2. Initialize the parameters of the model using
       :code:`Model.random_subset_initialization`.
    3. Perform full training.
    4. Save the inferred parameters (mode mixing coefficients, means and
       covariances) if :code:`save_inf_params=True`.

    This function will create two directories:

    - :code:`<output_dir>/model`, which contains the trained model.
    - :code:`<output_dir>/inf_params`, which contains the inferred parameters.

    Parameters
    ----------
    data : osl_dynamics.data.Data
        Data object for training the model.
    output_dir : str
        Path to output directory.
    config_kwargs : dict
        Keyword arguments to pass to `dynemo.Config <https://osl-dynamics\
        .readthedocs.io/en/latest/autoapi/osl_dynamics/models/dynemo\
        /index.html#osl_dynamics.models.dynemo.Config>`_. Defaults to::

            {'n_channels': data.n_channels.
             'sequence_length': 200,
             'inference_n_units': 64,
             'inference_normalization': 'layer',
             'model_n_units': 64,
             'model_normalization': 'layer',
             'learn_alpha_temperature': True,
             'initial_alpha_temperature': 1.0,
             'do_kl_annealing': True,
             'kl_annealing_curve': 'tanh',
             'kl_annealing_sharpness': 10,
             'n_kl_annealing_epochs': 20,
             'batch_size': 128,
             'learning_rate': 0.01,
             'lr_decay': 0.1,
             'n_epochs': 40}
    init_kwargs : dict, optional
        Keyword arguments to pass to :code:`Model.random_subset_initialization`.
        Defaults to::

            {'n_init': 5, 'n_epochs': 2, 'take': 1}.
    fit_kwargs : dict, optional
        Keyword arguments to pass to the :code:`Model.fit`.
    save_inf_params : bool, optional
        Should we save the inferred parameters?
    """

    init_kwargs = {} if init_kwargs is None else init_kwargs
    fit_kwargs = {} if fit_kwargs is None else fit_kwargs

    if data is None:
        raise ValueError("data must be passed.")

    from osl_dynamics.models import dynemo

    # Directories
    model_dir = output_dir + "/model"
    inf_params_dir = output_dir + "/inf_params"

    # Create the model object
    _logger.info("Building model")
    default_config_kwargs = {
        "n_channels": data.n_channels,
        "sequence_length": 200,
        "inference_n_units": 64,
        "inference_normalization": "layer",
        "model_n_units": 64,
        "model_normalization": "layer",
        "learn_alpha_temperature": True,
        "initial_alpha_temperature": 1.0,
        "do_kl_annealing": True,
        "kl_annealing_curve": "tanh",
        "kl_annealing_sharpness": 10,
        "n_kl_annealing_epochs": 20,
        "batch_size": 128,
        "learning_rate": 0.01,
        "lr_decay": 0.1,
        "n_epochs": 40,
    }
    config_kwargs = override_dict_defaults(default_config_kwargs, config_kwargs)
    _logger.info(f"Using config_kwargs: {config_kwargs}")
    config = dynemo.Config(**config_kwargs)
    model = dynemo.Model(config)
    model.summary()

    # Set regularisers
    model.set_regularizers(data)

    # Initialisation
    default_init_kwargs = {"n_init": 5, "n_epochs": 2, "take": 1}
    init_kwargs = override_dict_defaults(default_init_kwargs, init_kwargs)
    _logger.info(f"Using init_kwargs: {init_kwargs}")
    init_history = model.random_subset_initialization(data, **init_kwargs)

    # Keyword arguments for the fit method
    default_fit_kwargs = {}
    fit_kwargs = override_dict_defaults(default_fit_kwargs, fit_kwargs)
    _logger.info(f"Using fit_kwargs: {fit_kwargs}")

    # Training
    history = model.fit(data, **fit_kwargs)

    # Add free energy to the history object
    history["free_energy"] = history["loss"][-1]

    # Save trained model
    _logger.info(f"Saving model to: {model_dir}")
    model.save(model_dir)
    save(f"{model_dir}/init_history.pkl", init_history)
    save(f"{model_dir}/history.pkl", history)

    if save_inf_params:
        os.makedirs(inf_params_dir, exist_ok=True)

        # Get the inferred parameters
        alpha = model.get_alpha(data)
        means, covs = model.get_means_covariances()

        # Save inferred parameters
        save(f"{inf_params_dir}/alp.pkl", alpha)
        save(f"{inf_params_dir}/means.npy", means)
        save(f"{inf_params_dir}/covs.npy", covs)


def train_mdynemo(
    data,
    output_dir,
    config_kwargs,
    init_kwargs=None,
    fit_kwargs=None,
    corrs_init_kwargs=None,
    save_inf_params=True,
):
    """Train `MDyNeMo <https://osl-dynamics.readthedocs.io/en/latest/autoapi\
        /osl_dynamics/models/mdynemo/index.html>`_. This function will:
    
    1. Build an :code:`mdynemo.Model` object.
    2. Initialize the mode correlations using sliding window and KMeans.
    3. Initialize the parameters of the model using
        :code:`Model.random_subset_initialization`.
    4. Perform full training.
    5. Save the inferred parameters (mode time courses, means, stds and corrs)
        if :code:`save_inf_params=True`.

    This function will create two directories:

    - :code:`<output_dir>/model`, which contains the trained model.
    - :code:`<output_dir>/inf_params`, which contains the inferred parameters.

    Parameters
    ----------
    data : osl_dynamics.data.Data
        Data object for training the model.
    output_dir : str
        Path to output directory.
    config_kwargs : dict
        Keyword arguments to pass to `mdynemo.Config <https://osl-dynamics\
        .readthedocs.io/en/latest/autoapi/osl_dynamics/models/mdynemo\
        /index.html#osl_dynamics.models.mdynemo.Config>`_. Defaults to::

            {
                'n_channels': data.n_channels,
                'sequence_length': 200,
                'inference_n_units': 64,
                'inference_normalization': 'layer',
                'model_n_units': 64,
                'model_normalization': 'layer',
                'do_kl_annealing': True,
                'kl_annealing_curve': 'tanh',
                'kl_annealing_sharpness': 10,
                'n_kl_annealing_epochs': 20,
                'batch_size': 128,
                'learning_rate': 0.01,
                'lr_decay': 0.1,
                'n_epochs': 40,
            }.
    init_kwargs : dict, optional
        Keyword arguments to pass to :code:`Model.random_subset_initialization`.
        Defaults to::

            {'n_init': 5, 'n_epochs': 5, 'take': 1}.
    fit_kwargs : dict, optional
        Keyword arguments to pass to the :code:`Model.fit`.
    corrs_init_kwargs : dict, optional
        Keyword arguments to pass to the mode correlations
        initialisation. Defaults to::

            {
                'window_length': data.sampling_frequency * 2,
                'step_size': data.sampling_frequency // 25,
                'random_state': None,
                'n_init': 'auto',
                'init': 'k-means++',
            }.
    save_inf_params : bool, optional
        Should we save the inferred parameters?
    """
    if data is None:
        raise ValueError("data must be passed.")

    from osl_dynamics.models import mdynemo
    from osl_dynamics.analysis import connectivity
    from sklearn.cluster import KMeans

    init_kwargs = {} if init_kwargs is None else init_kwargs
    fit_kwargs = {} if fit_kwargs is None else fit_kwargs
    corrs_init_kwargs = {} if corrs_init_kwargs is None else corrs_init_kwargs

    # Directories
    model_dir = output_dir + "/model"
    inf_params_dir = output_dir + "/inf_params"

    _logger.info("Building model")
    default_config_kwargs = {
        "n_channels": data.n_channels,
        "sequence_length": 200,
        "inference_n_units": 64,
        "inference_normalization": "layer",
        "model_n_units": 64,
        "model_normalization": "layer",
        "do_kl_annealing": True,
        "kl_annealing_curve": "tanh",
        "kl_annealing_sharpness": 10,
        "n_kl_annealing_epochs": 20,
        "batch_size": 128,
        "learning_rate": 0.01,
        "lr_decay": 0.1,
        "n_epochs": 40,
    }
    config_kwargs = override_dict_defaults(default_config_kwargs, config_kwargs)
    _logger.info(f"Using config_kwargs: {config_kwargs}")
    config = mdynemo.Config(**config_kwargs)
    config.pca_components = data.pca_components

    # KMeans to intialise corrs
    _logger.info("Initialising corrs")
    default_corrs_init_kwargs = {
        "window_length": data.sampling_frequency * 2,
        "step_size": data.sampling_frequency // 25,
        "random_state": None,
        "n_init": "auto",
        "init": "k-means++",
    }
    corrs_init_kwargs = override_dict_defaults(
        default_corrs_init_kwargs, corrs_init_kwargs
    )
    _logger.info(f"Using corrs_init_kwargs: {corrs_init_kwargs}")
    tv_corr = connectivity.sliding_window_connectivity(
        data.time_series(),
        window_length=corrs_init_kwargs["window_length"],
        step_size=corrs_init_kwargs["step_size"],
        conn_type="corr",
        concatenate=True,
        n_jobs=data.n_jobs,
    )
    tv_corr = np.reshape(tv_corr, (tv_corr.shape[0], -1))
    kmeans = KMeans(
        n_clusters=config.n_corr_modes,
        n_init=corrs_init_kwargs["n_init"],
        init=corrs_init_kwargs["init"],
        random_state=corrs_init_kwargs["random_state"],
    ).fit(tv_corr)
    initial_corrs = kmeans.cluster_centers_.reshape(
        config.n_corr_modes, data.n_channels, data.n_channels
    )
    initial_corrs = array_ops.cov2corr(initial_corrs)
    config.initial_corrs = (
        config.pca_components @ initial_corrs @ config.pca_components.T
    )

    model = mdynemo.Model(config)
    model.summary()

    # Initialisation
    default_init_kwargs = {"n_init": 5, "n_epochs": 5, "take": 1}
    init_kwargs = override_dict_defaults(default_init_kwargs, init_kwargs)
    _logger.info(f"Using init_kwargs: {init_kwargs}")
    init_history = model.random_subset_initialization(data, **init_kwargs)

    # Keyword arguments for the fit method
    default_fit_kwargs = {}
    fit_kwargs = override_dict_defaults(default_fit_kwargs, fit_kwargs)
    _logger.info(f"Using fit_kwargs: {fit_kwargs}")

    # Training
    history = model.fit(data, **fit_kwargs)

    # Add free energy to the history object
    history["free_energy"] = model.free_energy(data)

    # Save trained model
    model.save(model_dir)
    save(f"{model_dir}/init_history.pkl", init_history)
    save(f"{model_dir}/history.pkl", history)

    if save_inf_params:
        del model
        model = mdynemo.Model.load(model_dir)
        os.makedirs(inf_params_dir, exist_ok=True)

        # Get the inferred parameters
        alpha, beta = model.get_mode_time_courses(data)
        means, stds, corrs = model.get_means_stds_corrs()

        save(f"{inf_params_dir}/alp.pkl", alpha)
        save(f"{inf_params_dir}/bet.pkl", beta)
        save(f"{inf_params_dir}/means.npy", means)
        save(f"{inf_params_dir}/stds.npy", stds)
        save(f"{inf_params_dir}/corrs.npy", corrs)


def train_hive(
    data,
    output_dir,
    config_kwargs,
    init_kwargs=None,
    fit_kwargs=None,
    save_inf_params=True,
):
    """ Train a `HIVE Model <https://osl-dynamics.\
    readthedocs.io/en/latest/autoapi/osl_dynamics/models/hive/index.html>`_.
    
    This function will:

    1. Build an :code:`hive.Model` object.
    2. Initialize the parameters of the HIVE model using
        :code:`Model.random_state_time_course_initialization`.
    3. Perform full training.
    4. Save the inferred parameters (state probabilities, means,
        covariances and embeddings) if :code:`save_inf_params=True`.
    
    This function will create two directories:

    - :code:`<output_dir>/model`, which contains the trained model.
    - :code:`<output_dir>/inf_params`, which contains the inferred parameters.
        This directory is only created if :code:`save_inf_params=True`.

    Parameters
    ----------
    data : osl_dynamics.data.Data
        Data object for training the model.
    output_dir : str
        Path to output directory.
    config_kwargs : dict
        Keyword arguments to pass to `hive.Config <https://osl-dynamics\
        .readthedocs.io/en/latest/autoapi/osl_dynamics/models/hive/index.html\
        #osl_dynamics.models.hive.Config>`_. Defaults to::

            {
                'sequence_length': 200,
                'spatial_embeddings_dim': 2,
                'dev_n_layers': 5,
                'dev_n_units': 32,
                'dev_activation': 'tanh',
                'dev_normalization': 'layer',
                'dev_regularizer': 'l1',
                'dev_regularizer_factor': 10,
                'batch_size': 128,
                'learning_rate': 0.005,
                'lr_decay': 0.1,
                'n_epochs': 30,
                'do_kl_annealing': True,
                'kl_annealing_curve': 'tanh',
                'kl_annealing_sharpness': 10,
                'n_kl_annealing_epochs': 15,
            }.
    init_kwargs : dict, optional
        Keyword arguments to pass to
        :code:`Model.random_state_time_course_initialization`. Defaults to::

            {'n_init': 10, 'n_epochs': 2}.
    fit_kwargs : dict, optional
        Keyword arguments to pass to the :code:`Model.fit`. No defaults.
    save_inf_params : bool, optional
        Should we save the inferred parameters?
    """
    if data is None:
        raise ValueError("data must be passed.")

    if not data.get_session_labels():
        data.add_session_labels("session_id", np.arange(data.n_sessions), "categorical")

    from osl_dynamics.models import hive

    init_kwargs = {} if init_kwargs is None else init_kwargs
    fit_kwargs = {} if fit_kwargs is None else fit_kwargs

    # Directories
    model_dir = output_dir + "/model"

    _logger.info("Building model")

    # SE-HMM config
    default_config_kwargs = {
        "n_channels": data.n_channels,
        "n_sessions": data.n_sessions,
        "sequence_length": 200,
        "spatial_embeddings_dim": 2,
        "dev_n_layers": 5,
        "dev_n_units": 32,
        "dev_activation": "tanh",
        "dev_normalization": "layer",
        "dev_regularizer": "l1",
        "dev_regularizer_factor": 10,
        "batch_size": 128,
        "learning_rate": 0.005,
        "lr_decay": 0.1,
        "n_epochs": 30,
        "do_kl_annealing": True,
        "kl_annealing_curve": "tanh",
        "kl_annealing_sharpness": 10,
        "n_kl_annealing_epochs": 15,
        "session_labels": data.get_session_labels(),
    }
    config_kwargs = override_dict_defaults(default_config_kwargs, config_kwargs)

    default_init_kwargs = {"n_init": 10, "n_epochs": 2}
    init_kwargs = override_dict_defaults(default_init_kwargs, init_kwargs)
    _logger.info(f"Using init_kwargs: {init_kwargs}")

    # Initialise and train HIVE
    _logger.info(f"Using config_kwargs: {config_kwargs}")
    config = hive.Config(**config_kwargs)
    model = hive.Model(config)
    model.summary()

    # Set regularisers
    model.set_regularizers(data)

    # Set deviation initializer
    model.set_dev_parameters_initializer(data)

    # Initialise HIVE
    _logger.info(f"Using init_kwargs: {init_kwargs}")
    init_history = model.random_state_time_course_initialization(
        data,
        **init_kwargs,
    )

    # Training
    history = model.fit(data, **fit_kwargs)

    _logger.info(f"Saving model to: {model_dir}")
    model.save(model_dir)

    del model
    model = hive.Model.load(model_dir)

    # Get the variational free energy
    history["free_energy"] = model.free_energy(data)

    # Save trained model
    save(f"{model_dir}/init_history.pkl", init_history)
    save(f"{model_dir}/history.pkl", history)

    if save_inf_params:
        # Make output directory
        inf_params_dir = output_dir + "/inf_params"
        os.makedirs(inf_params_dir, exist_ok=True)

        # Get the inferred parameters
        alpha = model.get_alpha(data)
        means, covs = model.get_means_covariances()
        session_means, session_covs = model.get_session_means_covariances()
        summed_embeddings = model.get_summed_embeddings()
        embedding_weights = model.get_embedding_weights()

        # Save inferred parameters
        save(f"{inf_params_dir}/alp.pkl", alpha)
        save(f"{inf_params_dir}/means.npy", means)
        save(f"{inf_params_dir}/covs.npy", covs)
        save(f"{inf_params_dir}/session_means.npy", session_means)
        save(f"{inf_params_dir}/session_covs.npy", session_covs)
        save(f"{inf_params_dir}/summed_embeddings.npy", summed_embeddings)
        save(f"{inf_params_dir}/embedding_weights.pkl", embedding_weights)


def get_inf_params(data, output_dir, observation_model_only=False):
    """Get inferred alphas.

    This function expects a model has already been trained and the following
    directory to exist:

    - :code:`<output_dir>/model`, which contains the trained model.

    This function will create the following directory:

    - :code:`<output_dir>/inf_params`, which contains the inferred parameters.

    Parameters
    ----------
    data : osl_dynamics.data.Data
        Data object.
    output_dir : str
        Path to output directory.
    observation_model_only : bool, optional
        We we only want to get the observation model parameters?
    """
    # Make output directory
    inf_params_dir = output_dir + "/inf_params"
    os.makedirs(inf_params_dir, exist_ok=True)

    #  Load model
    from osl_dynamics.models import load

    model_dir = output_dir + "/model"
    model = load(model_dir)

    if observation_model_only:
        # Get the inferred parameters
        means, covs = model.get_means_covariances()

        # Save
        save(f"{inf_params_dir}/means.npy", means)
        save(f"{inf_params_dir}/covs.npy", covs)
    else:
        # Get the inferred parameters
        alpha = model.get_alpha(data)
        means, covs = model.get_means_covariances()

        # Save
        save(f"{inf_params_dir}/alp.pkl", alpha)
        save(f"{inf_params_dir}/means.npy", means)
        save(f"{inf_params_dir}/covs.npy", covs)


def plot_power_maps_from_covariances(
    data,
    output_dir,
    mask_file=None,
    parcellation_file=None,
    power_save_kwargs=None,
):
    """Plot power maps calculated directly from the inferred covariances.

    This function expects a model has already been trained and the following
    directory to exist:

    - :code:`<output_dir>/inf_params`, which contains the inferred parameters.

    This function will output files called :code:`covs_.png` which contain
    plots of the power map of each state/mode taken directly from the inferred
    covariance matrices. The files will be saved to
    :code:`<output_dir>/inf_params`.

    This function also expects the data to be prepared in the same script
    that this wrapper is called from.

    Parameters
    ----------
    data : osl_dynamics.data.Data
        Data object.
    output_dir : str
        Path to output directory.
    mask_file : str, optional
        Mask file used to preprocess the training data. If :code:`None`,
        we use :code:`data.mask_file`.
    parcellation_file : str, optional
        Parcellation file used to parcellate the training data. If
        :code:`None`, we use :code:`data.parcellation_file`.
    power_save_kwargs : dict, optional
        Keyword arguments to pass to `analysis.power.save
        <https://osl-dynamics.readthedocs.io/en/latest/autoapi/osl_dynamics\
        /analysis/power/index.html#osl_dynamics.analysis.power.save>`_.
        Defaults to::

            {'filename': '<inf_params_dir>/covs_.png',
             'mask_file': data.mask_file,
             'parcellation_file': data.parcellation_file,
             'plot_kwargs': {'symmetric_cbar': True}}
    """
    # Validation
    power_save_kwargs = {} if power_save_kwargs is None else power_save_kwargs

    if mask_file is None:
        if data is None or data.mask_file is None:
            raise ValueError(
                "mask_file must be passed or specified in the Data object."
            )
        else:
            mask_file = data.mask_file

    if parcellation_file is None:
        if data is None or data.parcellation_file is None:
            raise ValueError(
                "parcellation_file must be passed or specified in the Data object."
            )
        else:
            parcellation_file = data.parcellation_file

    if hasattr(data, "n_embeddings"):
        n_embeddings = data.n_embeddings
    else:
        n_embeddings = 1

    if hasattr(data, "pca_components"):
        pca_components = data.pca_components
    else:
        pca_components = None

    # Directories
    inf_params_dir = f"{output_dir}/inf_params"

    # Load inferred covariances
    covs = load(f"{inf_params_dir}/covs.npy")

    # Reverse the effects of preparing the data
    from osl_dynamics.analysis import modes

    covs = modes.raw_covariances(covs, n_embeddings, pca_components)

    # Save
    from osl_dynamics.analysis import power

    default_power_save_kwargs = {
        "filename": f"{inf_params_dir}/covs_.png",
        "mask_file": mask_file,
        "parcellation_file": parcellation_file,
        "plot_kwargs": {"symmetric_cbar": True},
    }
    if "plot_kwargs" in power_save_kwargs:
        power_save_kwargs["plot_kwargs"] = override_dict_defaults(
            default_power_save_kwargs["plot_kwargs"],
            power_save_kwargs["plot_kwargs"],
        )
    power_save_kwargs = override_dict_defaults(
        default_power_save_kwargs, power_save_kwargs
    )
    _logger.info(f"Using power_save_kwargs: {power_save_kwargs}")
    power.save(covs, **power_save_kwargs)


def plot_tde_covariances(data, output_dir):
    """Plot inferred covariance of the time-delay embedded data.

    This function expects a model has already been trained and the following
    directory to exist:

    - :code:`<output_dir>/inf_params`, which contains the inferred parameters.

    This function will output a :code:`tde_covs.png` file containing a plot of
    the covariances in the :code:`<output_dir>/inf_params` directory.

    Parameters
    ----------
    data : osl_dynamics.data.Data
        Data object.
    output_dir : str
        Path to output directory.
    """
    inf_params_dir = f"{output_dir}/inf_params"

    covs = load(f"{inf_params_dir}/covs.npy")

    if hasattr(data, "pca_components"):
        if data.pca_components is not None:
            from osl_dynamics.analysis import modes

            covs = modes.reverse_pca(covs, data.pca_components)

    from osl_dynamics.utils import plotting

    plotting.plot_matrices(covs, filename=f"{inf_params_dir}/tde_covs.png")


def plot_state_psds(data, output_dir):
    """Plot state PSDs.

    This function expects multitaper spectra to have already been calculated
    and are in:

    - :code:`<output_dir>/spectra`.

    This function will output a file called :code:`psds.png` which contains
    a plot of each state PSD.

    Parameters
    ----------
    data : osl_dynamics.data.Data
        Data object.
    output_dir : str
        Path to output directory.
    """
    spectra_dir = f"{output_dir}/spectra"

    f = load(f"{spectra_dir}/f.npy")
    psd = load(f"{spectra_dir}/psd.npy")
    psd = np.mean(psd, axis=(0, 2))  # average over arrays and channels
    n_states = psd.shape[0]

    from osl_dynamics.utils import plotting

    plotting.plot_line(
        [f] * n_states,
        psd,
        labels=[f"State {i + 1}" for i in range(n_states)],
        x_label="Frequency (Hz)",
        y_label="PSD (a.u.)",
        x_range=[f[0], f[-1]],
        filename=f"{spectra_dir}/psds.png",
    )


def dual_estimation(data, output_dir, n_jobs=1):
    """Dual estimation for session-specific observation model parameters.

    This function expects a model has already been trained and the following
    directories to exist:

    - :code:`<output_dir>/model`, which contains the trained model.
    - :code:`<output_dir>/inf_params`, which contains the inferred parameters.

    This function will create the following directory:

    - :code:`<output_dir>/dual_estimates`, which contains the session-specific
      means and covariances.

    Parameters
    ----------
    data : osl_dynamics.data.Data
        Data object.
    output_dir : str
        Path to output directory.
    n_jobs : int, optional
        Number of jobs to run in parallel.
    """
    if data is None:
        raise ValueError("data must be passed.")

    # Directories
    model_dir = f"{output_dir}/model"
    inf_params_dir = f"{output_dir}/inf_params"
    dual_estimates_dir = f"{output_dir}/dual_estimates"
    os.makedirs(dual_estimates_dir, exist_ok=True)

    #  Load model
    from osl_dynamics import models

    model = models.load(model_dir)

    # Load the inferred state probabilities
    alpha = load(f"{inf_params_dir}/alp.pkl")

    # Dual estimation
    means, covs = model.dual_estimation(data, alpha=alpha, n_jobs=n_jobs)

    # Save
    save(f"{dual_estimates_dir}/means.npy", means)
    save(f"{dual_estimates_dir}/covs.npy", covs)


def multitaper_spectra(data, output_dir, kwargs, nnmf_components=None):
    """Calculate multitaper spectra.

    This function expects a model has already been trained and the following
    directories exist:

    - :code:`<output_dir>/model`, which contains the trained model.
    - :code:`<output_dir>/inf_params`, which contains the inferred parameters.

    This function will create the following directory:

    - :code:`<output_dir>/spectra`, which contains the post-hoc spectra.

    Parameters
    ----------
    data : osl_dynamics.data.Data
        Data object.
    output_dir : str
        Path to output directory.
    kwargs : dict
        Keyword arguments to pass to `analysis.spectral.multitaper_spectra
        <https://osl-dynamics.readthedocs.io/en/latest/autoapi/osl_dynamics\
        /analysis/spectral/index.html#osl_dynamics.analysis.spectral\
        .multitaper_spectra>`_. Defaults to::

            {'sampling_frequency': data.sampling_frequency,
             'keepdims': True}
    nnmf_components : int, optional
        Number of non-negative matrix factorization (NNMF) components to fit to
        the stacked session-specific coherence spectra.
    """
    if data is None:
        raise ValueError("data must be passed.")

    sampling_frequency = kwargs.pop("sampling_frequency", None)
    if sampling_frequency is None and data.sampling_frequency is None:
        raise ValueError(
            "sampling_frequency must be passed or specified in the Data object."
        )
    else:
        sampling_frequency = data.sampling_frequency

    default_kwargs = {
        "sampling_frequency": sampling_frequency,
        "keepdims": True,
    }
    kwargs = override_dict_defaults(default_kwargs, kwargs)
    _logger.info(f"Using kwargs: {kwargs}")

    # Directories
    model_dir = output_dir + "/model"
    inf_params_dir = output_dir + "/inf_params"
    spectra_dir = output_dir + "/spectra"
    os.makedirs(spectra_dir, exist_ok=True)

    # Load the inferred state probabilities
    alpha = load(f"{inf_params_dir}/alp.pkl")

    # Get the config used to create the model
    from osl_dynamics.models.mod_base import ModelBase

    model_config, _ = ModelBase.load_config(model_dir)

    # Get unprepared data (i.e. the data before calling Data.prepare)
    # We also trim the data to account for the data points lost to
    # time embedding or applying a sliding window
    data = data.trim_time_series(
        sequence_length=model_config["sequence_length"], prepared=False
    )

    # Calculate multitaper
    from osl_dynamics.analysis import spectral

    spectra = spectral.multitaper_spectra(data=data, alpha=alpha, **kwargs)

    # Unpack spectra and save
    return_weights = kwargs.pop("return_weights", False)
    if return_weights:
        f, psd, coh, w = spectra
        save(f"{spectra_dir}/f.npy", f)
        save(f"{spectra_dir}/psd.npy", psd)
        save(f"{spectra_dir}/coh.npy", coh)
        save(f"{spectra_dir}/w.npy", w)
    else:
        f, psd, coh = spectra
        save(f"{spectra_dir}/f.npy", f)
        save(f"{spectra_dir}/psd.npy", psd)
        save(f"{spectra_dir}/coh.npy", coh)

    if nnmf_components is not None:
        # Calculate NNMF and save
        nnmf = spectral.decompose_spectra(coh, n_components=nnmf_components)
        save(f"{spectra_dir}/nnmf_{nnmf_components}.npy", nnmf)


def nnmf(data, output_dir, n_components):
    """Calculate non-negative matrix factorization (NNMF).

    This function expects spectra have already been calculated and are in:

    - :code:`<output_dir>/spectra`, which contains multitaper spectra.

    Parameters
    ----------
    data : osl_dynamics.data.Data
        Data object.
    output_dir : str
        Path to output directory.
    n_components : int
        Number of components to fit.
    """
    from osl_dynamics.analysis import spectral

    spectra_dir = output_dir + "/spectra"
    coh = load(f"{spectra_dir}/coh.npy")
    nnmf = spectral.decompose_spectra(coh, n_components=n_components)
    save(f"{spectra_dir}/nnmf_{n_components}.npy", nnmf)


def regression_spectra(data, output_dir, kwargs):
    """Calculate regression spectra.

    This function expects a model has already been trained and the following
    directories exist:

    - :code:`<output_dir>/model`, which contains the trained model.
    - :code:`<output_dir>/inf_params`, which contains the inferred parameters.

    This function will create the following directory:

    - :code:`<output_dir>/spectra`, which contains the post-hoc spectra.

    Parameters
    ----------
    data : osl_dynamics.data.Data
        Data object.
    output_dir : str
        Path to output directory.
    kwargs : dict
        Keyword arguments to pass to `analysis.spectral.regress_spectra
        <https://osl-dynamics.readthedocs.io/en/latest/autoapi/osl_dynamics\
        /analysis/spectral/index.html#osl_dynamics.analysis.spectral\
        .regression_spectra>`_. Defaults to::

            {'sampling_frequency': data.sampling_frequency,
             'window_length': 4 * sampling_frequency,
             'step_size': 20,
             'n_sub_windows': 8,
             'return_coef_int': True,
             'keepdims': True}
    """
    if data is None:
        raise ValueError("data must be passed.")

    sampling_frequency = kwargs.pop("sampling_frequency", None)
    if sampling_frequency is None and data.sampling_frequency is None:
        raise ValueError(
            "sampling_frequency must be passed or specified in the Data object."
        )
    else:
        sampling_frequency = data.sampling_frequency

    default_kwargs = {
        "sampling_frequency": sampling_frequency,
        "window_length": int(4 * sampling_frequency),
        "step_size": 20,
        "n_sub_windows": 8,
        "return_coef_int": True,
        "keepdims": True,
    }
    kwargs = override_dict_defaults(default_kwargs, kwargs)
    _logger.info(f"Using kwargs: {kwargs}")

    # Directories
    model_dir = output_dir + "/model"
    inf_params_dir = output_dir + "/inf_params"
    spectra_dir = output_dir + "/spectra"
    os.makedirs(spectra_dir, exist_ok=True)

    # Load the inferred mixing coefficients
    alpha = load(f"{inf_params_dir}/alp.pkl")

    # Get the config used to create the model
    from osl_dynamics.models.mod_base import ModelBase

    model_config, _ = ModelBase.load_config(model_dir)

    # Get unprepared data (i.e. the data before calling Data.prepare)
    # We also trim the data to account for the data points lost to
    # time embedding or applying a sliding window
    data = data.trim_time_series(
        sequence_length=model_config["sequence_length"], prepared=False
    )

    # Calculate regression spectra
    from osl_dynamics.analysis import spectral

    spectra = spectral.regression_spectra(data=data, alpha=alpha, **kwargs)

    # Unpack spectra and save
    return_weights = kwargs.pop("return_weights", False)
    if return_weights:
        f, psd, coh, w = spectra
        save(f"{spectra_dir}/f.npy", f)
        save(f"{spectra_dir}/psd.npy", psd)
        save(f"{spectra_dir}/coh.npy", coh)
        save(f"{spectra_dir}/w.npy", w)
    else:
        f, psd, coh = spectra
        save(f"{spectra_dir}/f.npy", f)
        save(f"{spectra_dir}/psd.npy", psd)
        save(f"{spectra_dir}/coh.npy", coh)


def plot_group_ae_networks(
    data,
    output_dir,
    mask_file=None,
    parcellation_file=None,
    aec_abs=True,
    power_save_kwargs=None,
    conn_save_kwargs=None,
):
    """Plot group-level amplitude envelope networks.

    This function expects a model has been trained and the following directory
    to exist:

    - :code:`<output_dir>/inf_params`, which contains the inferred parameters.

    This function will create:

    - :code:`<output_dir>/networks`, which contains plots of the networks.

    Parameters
    ----------
    data : osl_dynamics.data.Data
        Data object.
    output_dir : str
        Path to output directory.
    mask_file : str, optional
        Mask file used to preprocess the training data. If :code:`None`,
        we use :code:`data.mask_file`.
    parcellation_file : str, optional
        Parcellation file used to parcellate the training data. If
        :code:`None`, we use :code:`data.parcellation_file`.
    aec_abs : bool, optional
        Should we take the absolute value of the amplitude envelope
        correlations?
    power_save_kwargs : dict, optional
        Keyword arguments to pass to `analysis.power.save
        <https://osl-dynamics.readthedocs.io/en/latest/autoapi/osl_dynamics\
        /analysis/power/index.html#osl_dynamics.analysis.power.save>`_.
        Defaults to::

            {'filename': '<output_dir>/networks/mean_.png',
             'mask_file': data.mask_file,
             'parcellation_file': data.parcellation_file}
    conn_save_kwargs : dict, optional
        Keyword arguments to pass to `analysis.connectivity.save
        <https://osl-dynamics.readthedocs.io/en/latest/autoapi/osl_dynamics\
        /analysis/connectivity/index.html#osl_dynamics.analysis.connectivity\
        .save>`_. Defaults to::

            {'parcellation_file': parcellation_file,
             'filename': '<output_dir>/networks/aec_.png',
             'threshold': 0.97,
             'plot_kwargs': {'display_mode': 'xz', 'annotate': False}}
    """
    power_save_kwargs = {} if power_save_kwargs is None else power_save_kwargs
    conn_save_kwargs = {} if conn_save_kwargs is None else conn_save_kwargs

    # Validation
    if mask_file is None:
        if data is None or data.mask_file is None:
            raise ValueError(
                "mask_file must be passed or specified in the Data object."
            )
        else:
            mask_file = data.mask_file

    if parcellation_file is None:
        if data is None or data.parcellation_file is None:
            raise ValueError(
                "parcellation_file must be passed or specified in the Data object."
            )
        else:
            parcellation_file = data.parcellation_file

    # Directories
    inf_params_dir = output_dir + "/inf_params"
    networks_dir = output_dir + "/networks"
    os.makedirs(networks_dir, exist_ok=True)

    # Load inferred means and covariances
    means = load(f"{inf_params_dir}/means.npy")
    covs = load(f"{inf_params_dir}/covs.npy")
    aecs = array_ops.cov2corr(covs)
    if aec_abs:
        aecs = abs(aecs)

    # Save mean activity maps
    from osl_dynamics.analysis import power

    default_power_save_kwargs = {
        "filename": f"{networks_dir}/mean_.png",
        "mask_file": mask_file,
        "parcellation_file": parcellation_file,
    }
    if "plot_kwargs" in power_save_kwargs:
        power_save_kwargs["plot_kwargs"] = override_dict_defaults(
            default_power_save_kwargs["plot_kwargs"],
            power_save_kwargs["plot_kwargs"],
        )
    power_save_kwargs = override_dict_defaults(
        default_power_save_kwargs, power_save_kwargs
    )
    _logger.info(f"Using power_save_kwargs: {power_save_kwargs}")
    power.save(means, **power_save_kwargs)

    # Save AEC networks
    from osl_dynamics.analysis import connectivity

    default_conn_save_kwargs = {
        "parcellation_file": parcellation_file,
        "filename": f"{networks_dir}/aec_.png",
        "threshold": 0.97,
        "plot_kwargs": {"display_mode": "xz", "annotate": False},
    }
    conn_save_kwargs = override_dict_defaults(
        default_conn_save_kwargs, conn_save_kwargs
    )
    _logger.info(f"Using conn_save_kwargs: {conn_save_kwargs}")
    connectivity.save(aecs, **conn_save_kwargs)


def plot_group_tde_hmm_networks(
    data,
    output_dir,
    mask_file=None,
    parcellation_file=None,
    frequency_range=None,
    percentile=97,
    power_save_kwargs=None,
    conn_save_kwargs=None,
):
    """Plot group-level TDE-HMM networks for a specified frequency band.

    This function will:

    1. Plot state PSDs.
    2. Plot the power maps.
    3. Plot coherence networks.

    This function expects spectra have already been calculated and are in:

    - :code:`<output_dir>/spectra`, which contains multitaper spectra.

    This function will create:

    - :code:`<output_dir>/networks`, which contains plots of the networks.

    Parameters
    ----------
    data : osl_dynamics.data.Data
        Data object.
    output_dir : str
        Path to output directory.
    mask_file : str, optional
        Mask file used to preprocess the training data. If :code:`None`,
        we use :code:`data.mask_file`.
    parcellation_file : str, optional
        Parcellation file used to parcellate the training data. If
        :code:`None`, we use :code:`data.parcellation_file`.
    frequency_range : list, optional
        List of length 2 containing the minimum and maximum frequency to
        integrate spectra over. Defaults to the full frequency range.
    percentile : float, optional
        Percentile for thresholding the coherence networks. Default is 97, which
        corresponds to the top 3% of edges (relative to the mean across states).
    power_save_kwargs : dict, optional
        Keyword arguments to pass to `analysis.power.save
        <https://osl-dynamics.readthedocs.io/en/latest/autoapi/osl_dynamics\
        /analysis/power/index.html#osl_dynamics.analysis.power.save>`_.
        Defaults to::

            {'mask_file': mask_file,
             'parcellation_file': parcellation_file,
             'filename': '<output_dir>/networks/pow_.png',
             'subtract_mean': True}
    conn_save_kwargs : dict, optional
        Keyword arguments to pass to `analysis.connectivity.save
        <https://osl-dynamics.readthedocs.io/en/latest/autoapi/osl_dynamics\
        /analysis/connectivity/index.html#osl_dynamics.analysis.connectivity\
        .save>`_. Defaults to::

            {'parcellation_file': parcellation_file,
             'filename': '<output_dir>/networks/coh_.png',
             'plot_kwargs': {'display_mode': "xz", 'annotate': False}}
    """
    power_save_kwargs = {} if power_save_kwargs is None else power_save_kwargs
    conn_save_kwargs = {} if conn_save_kwargs is None else conn_save_kwargs

    # Validation
    if mask_file is None:
        if data is None or data.mask_file is None:
            raise ValueError(
                "mask_file must be passed or specified in the Data object."
            )
        else:
            mask_file = data.mask_file

    if parcellation_file is None:
        if data is None or data.parcellation_file is None:
            raise ValueError(
                "parcellation_file must be passed or specified in the Data object."
            )
        else:
            parcellation_file = data.parcellation_file

    # Directories
    spectra_dir = output_dir + "/spectra"
    networks_dir = output_dir + "/networks"
    os.makedirs(networks_dir, exist_ok=True)

    # Load spectra
    f = load(f"{spectra_dir}/f.npy")
    psd = load(f"{spectra_dir}/psd.npy")
    coh = load(f"{spectra_dir}/coh.npy")
    if Path(f"{spectra_dir}/w.npy").exists():
        w = load(f"{spectra_dir}/w.npy")
    else:
        w = None
    if frequency_range is None:
        frequency_range = [f[0], f[-1]]

    # Calculate group average
    gpsd = np.average(psd, axis=0, weights=w)
    gcoh = np.average(coh, axis=0, weights=w)

    # Calculate group PSD averaged over channels for each state
    mgpsd = np.mean(gpsd, axis=0)
    p = np.mean(gpsd, axis=1)
    mp = np.mean(mgpsd, axis=0)

    # Plot PSDs
    from osl_dynamics.utils import plotting

    cmap = plt.get_cmap("tab10")
    for i in range(p.shape[0]):
        fig, ax = plotting.plot_line(
            [f],
            [mp],
            x_label="Frequency (Hz)",
            y_label="PSD (a.u.)",
            x_range=frequency_range,
            y_range=[0, np.max(p) * 1.1],
            plot_kwargs={"color": "black", "linestyle": "--"},
        )
        ax.plot(f, p[i], color=cmap(i))
        plotting.save(fig, filename=f"{networks_dir}/psd_{i}.png")

    # Calculate power maps from the group-level PSDs
    from osl_dynamics.analysis import power

    gp = power.variance_from_spectra(f, gpsd, frequency_range=frequency_range)

    # Save power maps
    default_power_save_kwargs = {
        "mask_file": mask_file,
        "parcellation_file": parcellation_file,
        "filename": f"{networks_dir}/pow_.png",
        "subtract_mean": True,
    }
    if "plot_kwargs" in power_save_kwargs:
        power_save_kwargs["plot_kwargs"] = override_dict_defaults(
            default_power_save_kwargs["plot_kwargs"],
            power_save_kwargs["plot_kwargs"],
        )
    power_save_kwargs = override_dict_defaults(
        default_power_save_kwargs, power_save_kwargs
    )
    _logger.info(f"Using power_save_kwargs: {power_save_kwargs}")
    power.save(gp, **power_save_kwargs)

    # Calculate coherence networks from group-level spectra
    from osl_dynamics.analysis import connectivity

    # Calculate group coherence relative to the average across states
    gc = connectivity.mean_coherence_from_spectra(
        f, gcoh, frequency_range=frequency_range
    )
    gc -= np.mean(gc, axis=0, keepdims=True)

    # Threshold
    gc = connectivity.threshold(gc, percentile=percentile, absolute_value=True)

    # Save coherence networks
    default_conn_save_kwargs = {
        "parcellation_file": parcellation_file,
        "filename": f"{networks_dir}/coh_.png",
        "plot_kwargs": {"display_mode": "xz", "annotate": False},
    }
    conn_save_kwargs = override_dict_defaults(
        default_conn_save_kwargs, conn_save_kwargs
    )
    _logger.info(f"Using conn_save_kwargs: {conn_save_kwargs}")
    connectivity.save(gc, **conn_save_kwargs)


def plot_group_nnmf_tde_hmm_networks(
    data,
    output_dir,
    nnmf_file,
    mask_file=None,
    parcellation_file=None,
    component=0,
    percentile=97,
    power_save_kwargs=None,
    conn_save_kwargs=None,
):
    """Plot group-level TDE-HMM networks using a NNMF component to integrate
    the spectra.

    This function will:

    1. Plot state PSDs.
    2. Plot the power maps.
    3. Plot coherence networks.

    This function expects spectra have already been calculated and are in:

    - :code:`<output_dir>/spectra`, which contains multitaper spectra.

    This function will create:

    - :code:`<output_dir>/networks`, which contains plots of the networks.

    Parameters
    ----------
    data : osl_dynamics.data.Data
        Data object.
    output_dir : str
        Path to output directory.
    nnmf_file : str
        Path relative to :code:`output_dir` for a npy file (with the output of
        `analysis.spectral.decompose_spectra
        <https://osl-dynamics.readthedocs.io/en/latest/autoapi/osl_dynamics\
        /analysis/spectral/index.html#osl_dynamics.analysis.spectral\
        .decompose_spectra>`_) containing the NNMF components.
    mask_file : str, optional
        Mask file used to preprocess the training data. If :code:`None`,
        we use :code:`data.mask_file`.
    parcellation_file : str, optional
        Parcellation file used to parcellate the training data. If
        :code:`None`, we use :code:`data.parcellation_file`.
    component : int, optional
        NNMF component to plot. Defaults to the first component.
    percentile : float, optional
        Percentile for thresholding the coherence networks. Default is 97, which
        corresponds to the top 3% of edges (relative to the mean across states).
    power_save_kwargs : dict, optional
        Keyword arguments to pass to `analysis.power.save
        <https://osl-dynamics.readthedocs.io/en/latest/autoapi/osl_dynamics\
        /analysis/power/index.html#osl_dynamics.analysis.power.save>`_.
        Defaults to::

            {'mask_file': mask_file,
             'parcellation_file': parcellation_file,
             'component': component,
             'filename': '<output_dir>/networks/pow_.png',
             'subtract_mean': True}
    conn_save_kwargs : dict, optional
        Keyword arguments to pass to `analysis.connectivity.save
        <https://osl-dynamics.readthedocs.io/en/latest/autoapi/osl_dynamics\
        /analysis/connectivity/index.html#osl_dynamics.analysis.connectivity\
        .save>`_. Defaults to::

            {'parcellation_file': parcellation_file,
             'component': component,
             'filename': '<output_dir>/networks/coh_.png',
             'plot_kwargs': {'display_mode': "xz", 'annotate': False}}
    """
    power_save_kwargs = {} if power_save_kwargs is None else power_save_kwargs
    conn_save_kwargs = {} if conn_save_kwargs is None else conn_save_kwargs

    # Validation
    if mask_file is None:
        if data is None or data.mask_file is None:
            raise ValueError(
                "mask_file must be passed or specified in the Data object."
            )
        else:
            mask_file = data.mask_file

    if parcellation_file is None:
        if data is None or data.parcellation_file is None:
            raise ValueError(
                "parcellation_file must be passed or specified in the Data object."
            )
        else:
            parcellation_file = data.parcellation_file

    # Directories
    spectra_dir = output_dir + "/spectra"
    networks_dir = output_dir + "/networks"
    os.makedirs(networks_dir, exist_ok=True)

    # Load the NNMF components
    nnmf_file = output_dir + "/" + nnmf_file
    if Path(nnmf_file).exists():
        nnmf = load(nnmf_file)
    else:
        raise ValueError(f"{nnmf_file} not found.")

    # Load spectra
    f = load(f"{spectra_dir}/f.npy")
    psd = load(f"{spectra_dir}/psd.npy")
    coh = load(f"{spectra_dir}/coh.npy")
    if Path(f"{spectra_dir}/w.npy").exists():
        w = load(f"{spectra_dir}/w.npy")
    else:
        w = None

    # Plot the NNMF components
    from osl_dynamics.utils import plotting

    n_components = nnmf.shape[0]
    plotting.plot_line(
        [f] * n_components,
        nnmf,
        labels=[f"Component {i}" for i in range(n_components)],
        x_label="Frequency (Hz)",
        y_label="Weighting",
        filename=f"{networks_dir}/nnmf.png",
    )

    # Calculate group average
    gpsd = np.average(psd, axis=0, weights=w)
    gcoh = np.average(coh, axis=0, weights=w)

    # Calculate group PSD averaged over channels for each state
    mgpsd = np.mean(gpsd, axis=0)
    p = np.mean(gpsd, axis=1)
    mp = np.mean(mgpsd, axis=0)

    # Plot PSDs
    from osl_dynamics.utils import plotting

    cmap = plt.get_cmap("tab10")
    for i in range(p.shape[0]):
        fig, ax = plotting.plot_line(
            [f],
            [mp],
            x_label="Frequency (Hz)",
            y_label="PSD (a.u.)",
            x_range=[f[0], f[-1]],
            y_range=[0, np.max(p) * 1.1],
            plot_kwargs={"color": "black", "linestyle": "--"},
        )
        ax.plot(f, p[i], color=cmap(i))
        plotting.save(fig, filename=f"{networks_dir}/psd_{i}.png")

    # Calculate power maps from the group-level PSDs
    from osl_dynamics.analysis import power

    gp = power.variance_from_spectra(f, gpsd, nnmf)

    # Save power maps
    default_power_save_kwargs = {
        "mask_file": mask_file,
        "parcellation_file": parcellation_file,
        "component": component,
        "filename": f"{networks_dir}/pow_.png",
        "subtract_mean": True,
    }
    if "plot_kwargs" in power_save_kwargs:
        power_save_kwargs["plot_kwargs"] = override_dict_defaults(
            default_power_save_kwargs["plot_kwargs"],
            power_save_kwargs["plot_kwargs"],
        )
    power_save_kwargs = override_dict_defaults(
        default_power_save_kwargs, power_save_kwargs
    )
    _logger.info(f"Using power_save_kwargs: {power_save_kwargs}")
    power.save(gp, **power_save_kwargs)

    # Calculate coherence networks from group-level spectra
    from osl_dynamics.analysis import connectivity

    # Calculate coherence networks from group-level spectra
    from osl_dynamics.analysis import connectivity

    # Calculate group coherence relative to the average across states
    gc = connectivity.mean_coherence_from_spectra(f, gcoh, nnmf)
    gc -= np.mean(gc, axis=0, keepdims=True)

    # Threshold
    gc = connectivity.threshold(gc, percentile=percentile, absolute_value=True)

    # Save coherence networks
    default_conn_save_kwargs = {
        "parcellation_file": parcellation_file,
        "component": component,
        "filename": f"{networks_dir}/coh_.png",
        "plot_kwargs": {"display_mode": "xz", "annotate": False},
    }
    conn_save_kwargs = override_dict_defaults(
        default_conn_save_kwargs, conn_save_kwargs
    )
    _logger.info(f"Using conn_save_kwargs: {conn_save_kwargs}")
    connectivity.save(gc, **conn_save_kwargs)


def plot_group_tde_dynemo_networks(
    data,
    output_dir,
    mask_file=None,
    parcellation_file=None,
    frequency_range=None,
    percentile=97,
    power_save_kwargs=None,
    conn_save_kwargs=None,
):
    """Plot group-level TDE-DyNeMo networks for a specified frequency band.

    This function will:

    1. Plot mode PSDs.
    2. Plot the power maps.
    3. Plot coherence networks.

    This function expects spectra have already been calculated and are in:

    - :code:`<output_dir>/spectra`, which contains regression spectra.

    This function will create:

    - :code:`<output_dir>/networks`, which contains plots of the networks.

    Parameters
    ----------
    data : osl_dynamics.data.Data
        Data object.
    output_dir : str
        Path to output directory.
    mask_file : str, optional
        Mask file used to preprocess the training data. If :code:`None`,
        we use :code:`data.mask_file`.
    parcellation_file : str, optional
        Parcellation file used to parcellate the training data. If
        :code:`None`, we use :code:`data.parcellation_file`.
    frequency_range : list, optional
        List of length 2 containing the minimum and maximum frequency to
        integrate spectra over. Defaults to the full frequency range.
    percentile : float, optional
        Percentile for thresholding the coherence networks. Default is 97, which
        corresponds to the top 3% of edges (relative to the mean across states).
    plot_save_kwargs : dict, optional
        Keyword arguments to pass to `analysis.power.save
        <https://osl-dynamics.readthedocs.io/en/latest/autoapi/osl_dynamics\
        /analysis/power/index.html#osl_dynamics.analysis.power.save>`_.
        Defaults to::

            {'mask_file': mask_file,
             'parcellation_file': parcellation_file,
             'filename': '<output_dir>/networks/pow_.png',
             'subtract_mean': True,
             'plot_kwargs': {'symmetric_cbar': True}}
    conn_save_kwargs : dict, optional
        Keyword arguments to pass to `analysis.connectivity.save
        <https://osl-dynamics.readthedocs.io/en/latest/autoapi/osl_dynamics\
        /analysis/connectivity/index.html#osl_dynamics.analysis.connectivity\
        .save>`_. Defaults to::

            {'parcellation_file': parcellation_file,
             'filename': '<output_dir>/networks/coh_.png',
             'plot_kwargs': {'edge_cmap': 'Reds'}}
    """
    power_save_kwargs = {} if power_save_kwargs is None else power_save_kwargs
    conn_save_kwargs = {} if conn_save_kwargs is None else conn_save_kwargs

    # Validation
    if mask_file is None:
        if data is None or data.mask_file is None:
            raise ValueError(
                "mask_file must be passed or specified in the Data object."
            )
        else:
            mask_file = data.mask_file

    if parcellation_file is None:
        if data is None or data.parcellation_file is None:
            raise ValueError(
                "parcellation_file must be passed or specified in the Data object."
            )
        else:
            parcellation_file = data.parcellation_file

    # Directories
    spectra_dir = output_dir + "/spectra"
    networks_dir = output_dir + "/networks"
    os.makedirs(networks_dir, exist_ok=True)

    # Load spectra
    f = load(f"{spectra_dir}/f.npy")
    psd = load(f"{spectra_dir}/psd.npy")
    coh = load(f"{spectra_dir}/coh.npy")
    if Path(f"{spectra_dir}/w.npy").exists():
        w = load(f"{spectra_dir}/w.npy")
    else:
        w = None

    # Only keep the regression coefficients
    psd = psd[:, 0]

    # Calculate group average
    gpsd = np.average(psd, axis=0, weights=w)
    gcoh = np.average(coh, axis=0, weights=w)

    # Calculate average PSD across channels and the standard error
    p = np.mean(gpsd, axis=-2)
    e = np.std(gpsd, axis=-2) / np.sqrt(gpsd.shape[-2])

    # Plot PSDs
    from osl_dynamics.utils import plotting

    n_modes = gpsd.shape[0]
    for i in range(n_modes):
        fig, ax = plotting.plot_line(
            [f],
            [p[i]],
            errors=[[p[i] - e[i]], [p[i] + e[i]]],
            labels=[f"Mode {i + 1}"],
            x_range=[f[0], f[-1]],
            y_range=[p.min() - 0.1 * p.max(), 1.4 * p.max()],
            x_label="Frequency (Hz)",
            y_label="PSD (a.u.)",
        )
        if frequency_range is not None:
            ax.axvspan(
                frequency_range[0],
                frequency_range[1],
                alpha=0.25,
                color="gray",
            )
        plotting.save(fig, filename=f"{networks_dir}/psd_{i}.png")

    # Calculate power maps from the group-level PSDs
    from osl_dynamics.analysis import power

    gp = power.variance_from_spectra(f, gpsd, frequency_range=frequency_range)

    # Save power maps
    default_power_save_kwargs = {
        "mask_file": mask_file,
        "parcellation_file": parcellation_file,
        "filename": f"{networks_dir}/pow_.png",
        "subtract_mean": True,
        "plot_kwargs": {"symmetric_cbar": True},
    }
    if "plot_kwargs" in power_save_kwargs:
        power_save_kwargs["plot_kwargs"] = override_dict_defaults(
            default_power_save_kwargs["plot_kwargs"],
            power_save_kwargs["plot_kwargs"],
        )
    power_save_kwargs = override_dict_defaults(
        default_power_save_kwargs, power_save_kwargs
    )
    _logger.info(f"Using power_save_kwargs: {power_save_kwargs}")
    power.save(gp, **power_save_kwargs)

    # Calculate coherence networks from group-level spectra
    from osl_dynamics.analysis import connectivity

    gc = connectivity.mean_coherence_from_spectra(
        f, gcoh, frequency_range=frequency_range
    )

    # Threshold
    gc = connectivity.threshold(gc, percentile=percentile, subtract_mean=True)

    # Save coherence networks
    default_conn_save_kwargs = {
        "parcellation_file": parcellation_file,
        "filename": f"{networks_dir}/coh_.png",
        "plot_kwargs": {"edge_cmap": "Reds"},
    }
    conn_save_kwargs = override_dict_defaults(
        default_conn_save_kwargs, conn_save_kwargs
    )
    _logger.info(f"Using conn_save_kwargs: {conn_save_kwargs}")
    connectivity.save(gc, **conn_save_kwargs)


def plot_alpha(
    data,
    output_dir,
    session=0,
    normalize=False,
    sampling_frequency=None,
    kwargs=None,
):
    """Plot inferred alphas.

    This is a wrapper for `utils.plotting.plot_alpha
    <https://osl-dynamics.readthedocs.io/en/latest/autoapi/osl_dynamics/utils\
    /plotting/index.html#osl_dynamics.utils.plotting.plot_alpha>`_.

    This function expects a model has been trained and the following directory
    to exist:

    - :code:`<output_dir>/inf_params`, which contains the inferred parameters.

    This function will create:

    - :code:`<output_dir>/alphas`, which contains plots of the inferred alphas.

    Parameters
    ----------
    data : osl_dynamics.data.Data
        Data object.
    output_dir : str
        Path to output directory.
    session : int, optional
        Index for session to plot. If 'all' is passed we create a separate plot
        for each session.
    normalize : bool, optional
        Should we also plot the alphas normalized using the trace of the
        inferred covariance matrices? Useful if we are plotting the inferred
        alphas from DyNeMo.
    sampling_frequency : float, optional
        Sampling frequency in Hz. If :code:`None`, we see if it is
        present in :code:`data.sampling_frequency`.
    kwargs : dict, optional
        Keyword arguments to pass to `utils.plotting.plot_alpha
        <https://osl-dynamics.readthedocs.io/en/latest/autoapi/osl_dynamics\
        /utils/plotting/index.html#osl_dynamics.utils.plotting.plot_alpha>`_.
        Defaults to::

            {'sampling_frequency': data.sampling_frequency,
             'filename': '<output_dir>/alphas/alpha_*.png'}
    """
    if sampling_frequency is None and data is not None:
        sampling_frequency = data.sampling_frequency

    # Directories
    inf_params_dir = output_dir + "/inf_params"
    alphas_dir = output_dir + "/alphas"
    os.makedirs(alphas_dir, exist_ok=True)

    # Load inferred alphas
    alp = load(f"{inf_params_dir}/alp.pkl")
    if isinstance(alp, np.ndarray):
        alp = [alp]

    # Plot
    from osl_dynamics.utils import plotting

    default_kwargs = {
        "sampling_frequency": sampling_frequency,
        "filename": f"{alphas_dir}/alpha_*.png",
    }
    kwargs = override_dict_defaults(default_kwargs, kwargs)
    _logger.info(f"Using kwargs: {kwargs}")

    if session == "all":
        for i in range(len(alp)):
            kwargs["filename"] = f"{alphas_dir}/alpha_{i}.png"
            plotting.plot_alpha(alp[i], **kwargs)
    else:
        kwargs["filename"] = f"{alphas_dir}/alpha_{session}.png"
        plotting.plot_alpha(alp[session], **kwargs)

    if normalize:
        from osl_dynamics.inference import modes

        # Calculate normalised alphas
        covs = load(f"{inf_params_dir}/covs.npy")
        norm_alp = modes.reweight_alphas(alp, covs)

        # Plot
        if session == "all":
            for i in range(len(alp)):
                kwargs["filename"] = f"{alphas_dir}/norm_alpha_{i}.png"
                plotting.plot_alpha(norm_alp[i], **kwargs)
        else:
            kwargs["filename"] = f"{alphas_dir}/norm_alpha_{session}.png"
            plotting.plot_alpha(norm_alp[session], **kwargs)


def calc_gmm_alpha(data, output_dir, kwargs=None):
    """Binarize inferred alphas using a two-component GMM.

    This function expects a model has been trained and the following directory
    to exist:

    - :code:`<output_dir>/inf_params`, which contains the inferred parameters.

    This function will create the following file:

    - :code:`<output_dir>/inf_params/gmm_alp.pkl`, which contains the binarized
      alphas.

    Parameters
    ----------
    data : osl_dynamics.data.Data
        Data object.
    output_dir : str
        Path to output directory.
    kwargs : dict, optional
        Keyword arguments to pass to `inference.modes.gmm_time_courses
        <https://osl-dynamics.readthedocs.io/en/latest/autoapi/osl_dynamics\
        /inference/modes/index.html#osl_dynamics.inference.modes\
        .gmm_time_courses>`_.
    """
    kwargs = {} if kwargs is None else kwargs
    inf_params_dir = output_dir + "/inf_params"

    # Load inferred alphas
    alp_file = f"{inf_params_dir}/alp.pkl"
    if not Path(alp_file).exists():
        raise ValueError(f"{alp_file} missing.")
    alp = load(alp_file)

    # Binarise using a two-component GMM
    from osl_dynamics.inference import modes

    _logger.info(f"Using kwargs: {kwargs}")
    gmm_alp = modes.gmm_time_courses(alp, **kwargs)
    save(f"{inf_params_dir}/gmm_alp.pkl", gmm_alp)


def plot_hmm_network_summary_stats(
    data,
    output_dir,
    use_gmm_alpha=False,
    sampling_frequency=None,
    sns_kwargs=None,
):
    """Plot HMM summary statistics for networks as violin plots.

    This function will plot the distribution over sessions for the following
    summary statistics:

    - Fractional occupancy.
    - Mean lifetime (s).
    - Mean interval (s).
    - Switching rate (Hz).

    This function expects a model has been trained and the following directory
    to exist:

    - :code:`<output_dir>/inf_params`, which contains the inferred parameters.

    This function will create:

    - :code:`<output_dir>/summary_stats`, which contains plots of the summary
      statistics.

    The :code:`<output_dir>/summary_stats` directory will also contain numpy
    files with the summary statistics.

    Parameters
    ----------
    data : osl_dynamics.data.Data
        Data object.
    output_dir : str
        Path to output directory.
    use_gmm_alpha : bool, optional
        Should we use alphas binarised using a Gaussian mixture model?
        This function assumes :code:`calc_gmm_alpha` has been called and the
        file :code:`<output_dir>/inf_params/gmm_alp.pkl` exists.
    sampling_frequency : float, optional
        Sampling frequency in Hz. If :code:`None`, we use
        :code:`data.sampling_frequency`.
    sns_kwargs : dict, optional
        Arguments to pass to :code:`sns.violinplot()`.
    """
    if sampling_frequency is None:
        if data is None or data.sampling_frequency is None:
            raise ValueError(
                "sampling_frequency must be passed or specified in the Data object."
            )
        else:
            sampling_frequency = data.sampling_frequency

    # Directories
    inf_params_dir = output_dir + "/inf_params"
    summary_stats_dir = output_dir + "/summary_stats"
    os.makedirs(summary_stats_dir, exist_ok=True)

    from osl_dynamics.inference import modes

    if use_gmm_alpha:
        # Use alphas that were binarised using a GMM
        gmm_alp_file = f"{inf_params_dir}/gmm_alp.pkl"
        if Path(gmm_alp_file).exists():
            stc = load(gmm_alp_file)
        else:
            raise ValueError(f"{gmm_alp_file} missing.")

    else:
        # Load inferred alphas and hard classify
        alp = load(f"{inf_params_dir}/alp.pkl")
        if isinstance(alp, np.ndarray):
            raise ValueError(
                "We must train on multiple sessions to plot the distribution "
                "of summary statistics."
            )
        stc = modes.argmax_time_courses(alp)

    # Calculate summary stats
    fo = modes.fractional_occupancies(stc)
    lt = modes.mean_lifetimes(stc, sampling_frequency)
    intv = modes.mean_intervals(stc, sampling_frequency)
    sr = modes.switching_rates(stc, sampling_frequency)

    # Save summary stats
    save(f"{summary_stats_dir}/fo.npy", fo)
    save(f"{summary_stats_dir}/lt.npy", lt)
    save(f"{summary_stats_dir}/intv.npy", intv)
    save(f"{summary_stats_dir}/sr.npy", sr)

    # Plot
    from osl_dynamics.utils import plotting

    n_states = fo.shape[1]
    x = range(1, n_states + 1)
    plotting.plot_violin(
        fo.T,
        x=x,
        x_label="State",
        y_label="Fractional Occupancy",
        filename=f"{summary_stats_dir}/fo.png",
        sns_kwargs=sns_kwargs,
    )
    plotting.plot_violin(
        lt.T,
        x=x,
        x_label="State",
        y_label="Mean Lifetime (s)",
        filename=f"{summary_stats_dir}/lt.png",
        sns_kwargs=sns_kwargs,
    )
    plotting.plot_violin(
        intv.T,
        x=x,
        x_label="State",
        y_label="Mean Interval (s)",
        filename=f"{summary_stats_dir}/intv.png",
        sns_kwargs=sns_kwargs,
    )
    plotting.plot_violin(
        sr.T,
        x=x,
        x_label="State",
        y_label="Switching rate (Hz)",
        filename=f"{summary_stats_dir}/sr.png",
        sns_kwargs=sns_kwargs,
    )


def plot_dynemo_network_summary_stats(data, output_dir):
    """Plot DyNeMo summary statistics for networks as violin plots.

    This function will plot the distribution over sessions for the following
    summary statistics:

    - Mean (renormalised) mixing coefficients.
    - Standard deviation of (renormalised) mixing coefficients.

    This function expects a model has been trained and the following directories
    to exist:

    - :code:`<output_dir>/model`, which contains the trained model.
    - :code:`<output_dir>/inf_params`, which contains the inferred parameters.

    This function will create:

    - :code:`<output_dir>/summary_stats`, which contains plots of the summary
      statistics.

    The :code:`<output_dir>/summary_stats` directory will also contain numpy
    files with the summary statistics.

    Parameters
    ----------
    data : osl_dynamics.data.Data
        Data object.
    output_dir : str
        Path to output directory.
    """
    # Directories
    model_dir = output_dir + "/model"
    inf_params_dir = output_dir + "/inf_params"
    summary_stats_dir = output_dir + "/summary_stats"
    os.makedirs(summary_stats_dir, exist_ok=True)

    # Load inferred parameters
    alp = load(f"{inf_params_dir}/alp.pkl")
    if isinstance(alp, np.ndarray):
        raise ValueError(
            "We must train on multiple sessions to plot the distribution "
            "of summary statistics."
        )

    # Get the config used to create the model
    from osl_dynamics.models.mod_base import ModelBase

    config, _ = ModelBase.load_config(model_dir)

    # Renormalise (only if we are learning covariances)
    from osl_dynamics.inference import modes

    if config["learn_covariances"]:
        covs = load(f"{inf_params_dir}/covs.npy")
        alp = modes.reweight_alphas(alp, covs)

    # Calculate summary stats
    alp_mean = np.array([np.mean(a, axis=0) for a in alp])
    alp_std = np.array([np.std(a, axis=0) for a in alp])
    alp_corr = np.array([np.corrcoef(a, rowvar=False) for a in alp])
    for c in alp_corr:
        np.fill_diagonal(c, 0)  # remove diagonal to see the off-diagonals better

    # Save summary stats
    save(f"{summary_stats_dir}/alp_mean.npy", alp_mean)
    save(f"{summary_stats_dir}/alp_std.npy", alp_std)
    save(f"{summary_stats_dir}/alp_corr.npy", alp_corr)

    # Plot
    from osl_dynamics.utils import plotting

    n_modes = alp_mean.shape[1]
    x = range(1, n_modes + 1)
    plotting.plot_violin(
        alp_mean.T,
        x=x,
        x_label="Mode",
        y_label="Mean",
        filename=f"{summary_stats_dir}/alp_mean.png",
    )
    plotting.plot_violin(
        alp_std.T,
        x=x,
        x_label="Mode",
        y_label="Standard Deviation",
        filename=f"{summary_stats_dir}/alp_std.png",
    )
    plotting.plot_matrices(
        np.mean(alp_corr, axis=0), filename=f"{summary_stats_dir}/alp_corr.png"
    )


def compare_groups_hmm_summary_stats(
    data,
    output_dir,
    group2_indices,
    separate_tests=False,
    covariates=None,
    n_perm=1000,
    n_jobs=1,
    sampling_frequency=None,
):
    """Compare HMM summary statistics between two groups.

    This function expects a model has been trained and the following directory
    to exist:

    - :code:`<output_dir>/inf_params`, which contains the inferred parameters.

    This function will create:

    - :code:`<output_dir>/group_diff`, which contains the summary statistics
      and plots.

    Parameters
    ----------
    data : osl_dynamics.data.Data
        Data object.
    output_dir : str
        Path to output directory.
    group2_indices : np.ndarray or list
        Indices indicating which sessions belong to the second group.
    separate_tests : bool, optional
        Should we perform a maximum statistic permutation test for each summary
        statistic separately?
    covariates : str, optional
        Path to a pickle file containing a :code:`dict` with covariances. Each
        item in the :code:`dict` must be the covariate name and value for each
        session. The covariates will be loaded with::

            from osl_dynamics.utils.misc import load
            covariates = load("/path/to/file.pkl")

        Example covariates::

            covariates = {"age": [...], "sex": [...]}
    n_perm : int, optional
        Number of permutations.
    n_jobs : int, optional
        Number of jobs for parallel processing.
    sampling_frequency : float, optional
        Sampling frequency in Hz. If :code:`None`, we use
        :code:`data.sampling_frequency`.
    """
    if sampling_frequency is None:
        if data is None or data.sampling_frequency is None:
            raise ValueError(
                "sampling_frequency must be passed or specified in the Data object."
            )
        else:
            sampling_frequency = data.sampling_frequency

    # Directories
    inf_params_dir = output_dir + "/inf_params"
    group_diff_dir = output_dir + "/group_diff"
    os.makedirs(group_diff_dir, exist_ok=True)

    # Get inferred state time courses
    from osl_dynamics.inference import modes

    alp = load(f"{inf_params_dir}/alp.pkl")
    stc = modes.argmax_time_courses(alp)

    # Calculate summary stats
    names = ["fo", "lt", "intv", "sr"]
    fo = modes.fractional_occupancies(stc)
    lt = modes.mean_lifetimes(stc, sampling_frequency)
    intv = modes.mean_intervals(stc, sampling_frequency)
    sr = modes.switching_rates(stc, sampling_frequency)
    sum_stats = np.swapaxes([fo, lt, intv, sr], 0, 1)

    # Save
    for i in range(4):
        save(f"{group_diff_dir}/{names[i]}.npy", sum_stats[:, i])

    # Create a vector for group assignments
    n_sessions = fo.shape[0]
    assignments = np.ones(n_sessions)
    assignments[group2_indices] += 1

    # Load covariates
    if covariates is not None:
        covariates = load(covariates)
    else:
        covariates = {}

    # Perform statistical significance testing
    from osl_dynamics.analysis import statistics

    if separate_tests:
        pvalues = []
        for i in range(4):
            # Calculate a statistical significance test for each
            # summary stat separately
            _, p = statistics.group_diff_max_stat_perm(
                sum_stats[:, i],
                assignments,
                n_perm=n_perm,
                covariates=covariates,
                n_jobs=n_jobs,
            )
            pvalues.append(p)
            _logger.info(f"{names[i]}: {np.sum(p <  0.05)} states have p-value<0.05")
            save(f"{group_diff_dir}/{names[i]}_pvalues.npy", p)
        pvalues = np.array(pvalues)
    else:
        # Calculate a statistical significance test for all
        # summary stats concatenated
        _, pvalues = statistics.group_diff_max_stat_perm(
            sum_stats,
            assignments,
            n_perm=n_perm,
            covariates=covariates,
            n_jobs=n_jobs,
        )
        for i in range(4):
            _logger.info(
                f"{names[i]}: {np.sum(pvalues[i] < 0.05)} states have p-value<0.05"
            )
            save(f"{group_diff_dir}/{names[i]}_pvalues.npy", pvalues[i])

    # Plot
    from osl_dynamics.utils import plotting

    labels = [
        "Fractional Occupancy",
        "Mean Lifetime (s)",
        "Mean Interval (s)",
        "Switching Rate (Hz)",
    ]
    for i in range(4):
        plotting.plot_summary_stats_group_diff(
            name=labels[i],
            summary_stats=sum_stats[:, i],
            pvalues=pvalues[i],
            assignments=assignments,
            filename=f"{group_diff_dir}/{names[i]}.png",
        )


def plot_burst_summary_stats(data, output_dir, sampling_frequency=None):
    """Plot burst summary statistics as violin plots.

    This function will plot the distribution over sessions for the following
    summary statistics:

    - Mean lifetime (s).
    - Mean interval (s).
    - Burst count (Hz).
    - Mean amplitude (a.u.).

    This function expects a model has been trained and the following
    directories to exist:

    - :code:`<output_dir>/model`, which contains the trained model.
    - :code:`<output_dir>/inf_params`, which contains the inferred parameters.

    This function will create:

    - :code:`<output_dir>/summary_stats`, which contains plots of the summary
      statistics.

    The :code:`<output_dir>/summary_stats` directory will also contain numpy
    files with the summary statistics.

    Parameters
    ----------
    data : osl_dynamics.data.Data
        Data object.
    output_dir : str
        Path to output directory.
    sampling_frequency : float, optional
        Sampling frequency in Hz. If :code:`None`, we use
        :code:`data.sampling_frequency`.
    """
    if sampling_frequency is None:
        if data is None or data.sampling_frequency is None:
            raise ValueError(
                "sampling_frequency must be passed or specified in the Data object."
            )
        else:
            sampling_frequency = data.sampling_frequency

    # Directories
    model_dir = output_dir + "/model"
    inf_params_dir = output_dir + "/inf_params"
    summary_stats_dir = output_dir + "/summary_stats"
    os.makedirs(summary_stats_dir, exist_ok=True)

    from osl_dynamics.inference import modes

    # Load state time course
    alp = load(f"{inf_params_dir}/alp.pkl")
    stc = modes.argmax_time_courses(alp)

    # Get the config used to create the model
    from osl_dynamics.models.mod_base import ModelBase

    model_config, _ = ModelBase.load_config(model_dir)

    # Get unprepared data (i.e. the data before calling Data.prepare)
    # We also trim the data to account for the data points lost to
    # time embedding or applying a sliding window
    data = data.trim_time_series(
        sequence_length=model_config["sequence_length"], prepared=False
    )

    # Calculate summary stats
    lt = modes.mean_lifetimes(stc, sampling_frequency)
    intv = modes.mean_intervals(stc, sampling_frequency)
    bc = modes.switching_rates(stc, sampling_frequency)
    amp = modes.mean_amplitudes(stc, data)

    # Save summary stats
    save(f"{summary_stats_dir}/lt.npy", lt)
    save(f"{summary_stats_dir}/intv.npy", intv)
    save(f"{summary_stats_dir}/bc.npy", bc)
    save(f"{summary_stats_dir}/amp.npy", amp)

    from osl_dynamics.utils import plotting

    # Plot
    n_states = lt.shape[1]
    plotting.plot_violin(
        lt.T,
        x=range(1, n_states + 1),
        x_label="State",
        y_label="Mean Lifetime (s)",
        filename=f"{summary_stats_dir}/lt.png",
    )
    plotting.plot_violin(
        intv.T,
        x=range(1, n_states + 1),
        x_label="State",
        y_label="Mean Interval (s)",
        filename=f"{summary_stats_dir}/intv.png",
    )
    plotting.plot_violin(
        bc.T,
        x=range(1, n_states + 1),
        x_label="State",
        y_label="Burst Count (Hz)",
        filename=f"{summary_stats_dir}/bc.png",
    )
    plotting.plot_violin(
        amp.T,
        x=range(1, n_states + 1),
        x_label="State",
        y_label="Mean Amplitude (a.u.)",
        filename=f"{summary_stats_dir}/amp.png",
    )



"""Functions to calculate and plot network connectivity."""

import os
import logging
from pathlib import Path

import numpy as np
from nilearn import plotting
from scipy import stats
from tqdm.auto import trange
from pqdm.threads import pqdm
import matplotlib.pyplot as plt

from osl_dynamics import array_ops
from osl_dynamics.analysis import gmm
from osl_dynamics.analysis.spectral import get_frequency_args_range
from osl_dynamics.utils.misc import override_dict_defaults
from osl_dynamics.utils.parcellation import Parcellation

_logger = logging.getLogger("osl-dynamics")


def sliding_window_connectivity(
    data,
    window_length,
    step_size=None,
    conn_type="corr",
    concatenate=False,
    n_jobs=1,
):
    """Calculate sliding window connectivity.

    Parameters
    ----------
    data : list or np.ndarray
        Time series data. Shape must be (n_sessions, n_samples, n_channels)
        or (n_samples, n_channels).
    window_length : int
        Window length in samples.
    step_size : int, optional
        Number of samples to slide the window along the time series.
        If :code:`None`, then :code:`step_size=window_length // 2`.
    conn_type : str, optional
        Metric to use to calculate pairwise connectivity in the network.
        Should use :code:`"corr"` for Pearson correlation or :code:`"cov"`
        for covariance.
    concatenate : bool, optional
        Should we concatenate the sliding window connectivities from each
        array into one big time series?
    n_jobs : int, optional
        Number of parallel jobs to run. Default is 1.

    Returns
    -------
    sliding_window_conn : list or np.ndarray
        Time series of connectivity matrices. Shape is (n_sessions, n_windows,
        n_channels, n_channels) or (n_windows, n_channels, n_channels).
    """
    # Validation
    if conn_type not in ["corr", "cov"]:
        raise ValueError("conn_type must be 'corr' or 'cov'.")

    if conn_type == "cov":
        metric = np.cov
    else:
        metric = np.corrcoef

    if step_size is None:
        step_size = window_length // 2

    if isinstance(data, np.ndarray):
        if data.ndim != 3:
            data = [data]

    # Helper function to calculate connectivity
    def _swc(x):
        n_samples = x.shape[0]
        n_channels = x.shape[1]
        n_windows = (n_samples - window_length - 1) // step_size + 1

        # Preallocate an array to hold moving average values
        swc = np.empty([n_windows, n_channels, n_channels], dtype=np.float32)

        # Compute connectivity matrix for each window
        for i in range(n_windows):
            window_ts = x[i * step_size : i * step_size + window_length]
            swc[i] = metric(window_ts, rowvar=False)

        return swc

    # Setup keyword arguments to pass to the helper function
    kwargs = [{"x": x} for x in data]

    if len(data) == 1:
        _logger.info("Sliding window connectivity")
        results = [_swc(**kwargs[0])]

    elif n_jobs == 1:
        results = []
        for i in trange(len(data), desc="Sliding window connectivity"):
            results.append(_swc(**kwargs[i]))

    else:
        _logger.info("Sliding window connectivity")
        results = pqdm(
            kwargs,
            _swc,
            argument_type="kwargs",
            n_jobs=n_jobs,
        )

    if concatenate or len(results) == 1:
        results = np.concatenate(results)

    return results


def covariance_from_spectra(
    f,
    cpsd,
    components=None,
    frequency_range=None,
):
    """Calculates covariance from cross power spectra.

    Parameters
    ----------
    f : np.ndarray
        Frequency axis of the spectra. Only used if :code:`frequency_range`
        is given. Shape must be (n_freq,).
    cpsd : np.ndarray
        Cross power spectra. Shape must be
        (n_modes, n_channels, n_channels, n_freq).
    components : np.ndarray, optional
        Spectral components. Shape must be (n_components, n_freq).
    frequency_range : list, optional
        Frequency range to integrate the PSD over (Hz).
        Default is the full range.

    Returns
    -------
    covar : np.ndarray
        Covariance over a frequency band for each component of each mode.
        Shape is (n_components, n_modes, n_channels, n_channels).
    """

    # Validation
    error_message = (
        "A (n_channels, n_channels, n_freq), "
        + "(n_modes, n_channels, n_channels, n_freq) or "
        + "(n_sessions, n_modes, n_channels, n_channels, n_freq) "
        + "array must be passed."
    )
    cpsd = array_ops.validate(
        cpsd,
        correct_dimensionality=5,
        allow_dimensions=[3, 4],
        error_message=error_message,
    )

    if components is not None and frequency_range is not None:
        raise ValueError(
            "Only one of the arguments components or frequency range can be passed."
        )

    if frequency_range is not None and f is None:
        raise ValueError(
            "If frequency_range is passed, frequenices must also be passed."
        )

    # Dimensions
    n_sessions, n_modes, n_channels, n_channels, n_freq = cpsd.shape
    if components is None:
        n_components = 1
    else:
        n_components = components.shape[0]

    # Calculate connectivity maps for each array
    covar = []
    for i in range(n_sessions):
        # Cross spectral densities
        csd = cpsd[i].reshape(-1, n_freq)
        csd = abs(csd)

        if components is not None:
            # Calculate PSD for each spectral component
            c = components @ csd.T
            for j in range(n_components):
                c[j] /= np.sum(components[j])

        else:
            # Integrate over the given frequency range
            if frequency_range is None:
                c = np.sum(csd, axis=-1)
            else:
                [min_arg, max_arg] = get_frequency_args_range(f, frequency_range)
                c = np.sum(csd[..., min_arg:max_arg], axis=-1)

        c = c.reshape(n_components, n_modes, n_channels, n_channels)
        covar.append(c)

    return np.squeeze(covar)


def mean_coherence_from_spectra(
    f,
    coh,
    components=None,
    frequency_range=None,
):
    """Calculates mean coherence from spectra.

    Parameters
    ----------
    f : np.ndarray
        Frequency axis of the spectra. Only used if :code:`frequency_range` is
        given. Shape must be (n_freq,).
    coh : np.ndarray
        Coherence for each channel. Shape must be (n_modes, n_channels,
        n_channels, n_freq).
    components : np.ndarray, optional
        Spectral components. Shape must be (n_components, n_freq).
    frequency_range : list, optional
        Frequency range to integrate the PSD over (Hz).

    Returns
    -------
    mean_coh : np.ndarray
        Mean coherence over a frequency band for each component of each mode.
        Shape is (n_components, n_modes, n_channels, n_channels) or
        (n_modes, n_channels, n_channels) or (n_channels, n_channels).
    """

    # Validation
    error_message = (
        "a 3D numpy array (n_channels, n_channels, n_freq) "
        + "or 4D numpy array (n_modes, n_channels, n_channels, "
        + "n_freq) must be passed for spectra."
    )
    coh = array_ops.validate(
        coh,
        correct_dimensionality=5,
        allow_dimensions=[3, 4],
        error_message=error_message,
    )

    if components is not None and frequency_range is not None:
        raise ValueError(
            "Only one of the arguments components or frequency range can be passed."
        )

    if frequency_range is not None and f is None:
        raise ValueError(
            "If frequency_range is passed, frequenices must also be passed."
        )

    # Dimensions
    n_sessions, n_modes, n_channels, n_channels, n_freq = coh.shape
    if components is None:
        n_components = 1
    else:
        n_components = components.shape[0]

    # Calculate mean coherence for each array
    mean_coh = []
    for i in range(n_sessions):
        # Concatenate over modes
        c = coh[i].reshape(-1, n_freq)

        if components is not None:
            # Coherence for each spectral component
            c = components @ c.T
            for j in range(n_components):
                c[j] /= np.sum(components[j])

        else:
            # Mean over the given frequency range
            if frequency_range is None:
                c = np.mean(c, axis=-1)
            else:
                [min_arg, max_arg] = get_frequency_args_range(f, frequency_range)
                c = np.mean(c[..., min_arg:max_arg], axis=-1)

        c = c.reshape(n_components, n_modes, n_channels, n_channels)
        mean_coh.append(c)

    return np.squeeze(mean_coh)


def mean_connections(conn_map):
    """Average the edges for each node.

    Parameters
    ----------
    conn_map : np.ndarray
        A (..., n_channels, n_channels) connectivity matrix.

    Returns
    -------
    mean_connections : np.ndarray
        A (..., n_channels) matrix.
    """
    return np.mean(conn_map, axis=-1)


def eigenvectors(
    conn_map,
    n_eigenvectors=1,
    absolute_value=False,
    as_network=False,
):
    """Calculate eigenvectors of a connectivity matrix.

    Parameters
    ----------
    conn_map : np.ndarray
        Connectivity matrix. Shape must be (..., n_channels, n_channels).
    n_eigenvectors : int, optional
        Number of eigenvectors to include.
    absolute_value : bool, optional
        Should we take the absolute value of the connectivity matrix before
        calculating the eigen decomposition?
    as_network : bool, optional
        Should we return a matrix?

    Returns
    -------
    eigenvectors : np.ndarray.
        Eigenvectors. Shape is (n_eigenvectors, ..., n_channels, n_channels)
        if :code:`as_network=True`, otherwise it is
        (n_eigenvectors, ..., n_channels). If :code:`n_eigenvectors=1`,
        the first dimension is removed.
    """
    if absolute_value:
        # Take absolute value
        conn_map = abs(conn_map)

    # Calculate eigen decomposition
    _, eigenvectors = np.linalg.eigh(conn_map)

    # Reorder from ascending eigenvalues to descending
    eigenvectors = eigenvectors[..., ::-1]

    # Keep the requested number of eigenvectors and make the first axis
    # specify the eigenvector
    eigenvectors = np.rollaxis(eigenvectors[..., :n_eigenvectors], -1)

    if as_network:
        # Calculate the outer product using the eigenvectors
        eigenvectors = np.expand_dims(eigenvectors, axis=-1) @ np.expand_dims(
            eigenvectors, axis=-2
        )

    return np.squeeze(eigenvectors)


def gmm_threshold(
    conn_map,
    subtract_mean=False,
    mean_weights=None,
    standardize=False,
    p_value=None,
    keep_positive_only=False,
    one_component_percentile=0,
    n_sigma=0,
    sklearn_kwargs=None,
    show=False,
    filename=None,
    plot_kwargs=None,
):
    """Threshold a connectivity matrix using the GMM method.

    Wrapper for combining `connectivity.fit_gmm <https://osl-dynamics\
    .readthedocs.io/en/latest/autoapi/osl_dynamics/analysis/connectivity\
    /index.html#osl_dynamics.analysis.connectivity.fit_gmm>`_ and
    `connectivity.threshold <https://osl-dynamics.readthedocs.io/en/latest\
    /autoapi/osl_dynamics/analysis/connectivity/index.html#osl_dynamics\
    .analysis.connectivity.threshold>`_.

    Parameters
    ----------
    conn_map : np.ndarray
        Connectivity matrix. Shape must be (n_components, n_modes,
        n_channels, n_channels) or (n_modes, n_channels, n_channels)
        or (n_channels, n_channels).
    subtract_mean : bool, optional
        Should we subtract the mean over modes before fitting a GMM?
    mean_weights: np.ndarray, optional
        Numpy array with weightings for each mode/state to use to calculate
        the mean. Default is equal weighting.
    standardize : bool, optional
        Should we standardize the input to the GMM?
    p_value : float, optional
        Used to determine a threshold. We ensure the data points assigned to
        the 'on' component have a probability of less than :code:`p_value` of
        belonging to the 'off' component.
    keep_positive_only : bool, optional
        Should we only keep positive values to fit a GMM to?
    one_component_percentile : float, optional
        Percentile threshold if only one component is found. Should be a
        between 0 and 100. E.g. for the 95th percentile,
        :code:`one_component_percentile=95`.
    n_sigma : float, optional
        Number of standard deviations of the 'off' component the mean of the
        'on' component must be for the fit to be considered to have two
        components.
    sklearn_kwargs : dict, optional
        Dictionary of keyword arguments to pass to
        `sklearn.mixture.GaussianMixture <https://scikit-learn.org/stable\
        /modules/generated/sklearn.mixture.GaussianMixture.html>`_.
    show : bool, optional
        Should we show the GMM fit to the distribution of :code:`conn_map`.
    filename : str, optional
        Filename to save fit to.
    plot_kwargs : dict, optional
        Dictionary of keyword arguments to pass to `utils.plotting.plot_gmm
        <https://osl-dynamics.readthedocs.io/en/latest/autoapi/osl_dynamics\
        /utils/plotting/index.html#osl_dynamics.utils.plotting.plot_gmm>`_.

    Returns
    -------
    conn_map : np.ndarray
        Thresholded connectivity matrix. The shape is the same as the original
        :code:`conn_map`.
    """
    percentile = fit_gmm(
        conn_map,
        subtract_mean,
        mean_weights,
        standardize,
        p_value,
        keep_positive_only,
        one_component_percentile,
        n_sigma,
        sklearn_kwargs,
        show,
        filename,
        plot_kwargs,
    )
    conn_map = threshold(conn_map, percentile, subtract_mean)
    return conn_map


def fit_gmm(
    conn_map,
    subtract_mean=False,
    mean_weights=None,
    standardize=False,
    p_value=None,
    keep_positive_only=False,
    one_component_percentile=0,
    n_sigma=0,
    sklearn_kwargs=None,
    show=False,
    filename=None,
    plot_kwargs=None,
):
    """Fit a two component GMM to connections to identify a threshold.

    Parameters
    ----------
    conn_map : np.ndarray
        Connectivity map. Shape must be (n_components, n_modes, n_channels,
        n_channels) or (n_modes, n_channels, n_channels) or (n_channels,
        n_channels).
    subtract_mean : bool, optional
        Should we subtract the mean over modes before fitting a GMM?
    mean_weights: np.ndarray, optional
        Numpy array with weightings for each mode/state to use to calculate
        the mean. Default is equal weighting.
    standardize : bool, optional
        Should we standardize the input to the GMM?
    p_value : float, optional
        Used to determine a threshold. We ensure the data points assigned to
        the 'on' component have a probability of less than :code:`p_value` of
        belonging to the 'off' component.
    keep_positive_only : bool, optional
        Should we only keep positive values to fit a GMM to?
    one_component_percentile : float, optional
        Percentile threshold if only one component is found. Should be a
        between 0 and 100. E.g. for the 95th percentile,
        :code:`one_component_percentile=95`.
    n_sigma : float, optional
        Number of standard deviations of the 'off' component the mean of the
        'on' component must be for the fit to be considered to have two
        components.
    sklearn_kwargs : dict, optional
        Dictionary of keyword arguments to pass to
        `sklearn.mixture.GaussianMixture <https://scikit-learn.org/stable\
        /modules/generated/sklearn.mixture.GaussianMixture.html>`_
        Default is :code:`{"max_iter": 5000, "n_init": 10}`.
    show : bool, optional
        Should we show the GMM fit to the distribution of :code:`conn_map`.
    filename : str, optional
        Filename to save fit to.
    plot_kwargs : dict, optional
        Dictionary of keyword arguments to pass to `utils.plotting.plot_gmm
        <https://osl-dynamics.readthedocs.io/en/latest/autoapi/osl_dynamics\
        /utils/plotting/index.html#osl_dynamics.utils.plotting.plot_gmm>`_.

    Returns
    -------
    percentile : np.ndarray
        Percentile threshold. Shape is (n_components, n_modes) or (n_modes,).
    """
    # Validation
    conn_map = array_ops.validate(
        conn_map,
        correct_dimensionality=4,
        allow_dimensions=[2, 3],
        error_message="conn_map must be (n_modes, n_channels, n_channels) "
        + "or (n_channels, n_channels).",
    )

    if sklearn_kwargs is None:
        sklearn_kwargs = {}
    default_sklearn_kwargs = {"max_iter": 5000, "n_init": 10}
    sklearn_kwargs = override_dict_defaults(default_sklearn_kwargs, sklearn_kwargs)

    # Number of components, modes and channels
    n_components = conn_map.shape[0]
    n_modes = conn_map.shape[1]
    n_channels = conn_map.shape[2]

    # Mean over modes
    mean_conn_map = np.average(conn_map, axis=1, weights=mean_weights)

    # Indices for off diagonal elements
    m, n = np.triu_indices(n_channels, k=1)

    # Calculate thresholds by fitting a GMM
    percentiles = np.empty([n_components, n_modes])
    for i in range(n_components):
        for j in range(n_modes):
            # Off diagonal connectivity values to fit a GMM to
            if subtract_mean:
                c = conn_map[i, j, m, n] - mean_conn_map[i, m, n]
            else:
                c = conn_map[i, j, m, n]

            if keep_positive_only:
                # Only keep positive entries
                # (this is what's done in MATLAB OSL's teh_graph_gmm_fit)
                c = c[c > 0]
                if len(c) == 0:
                    percentiles[i, j] = 100
                    continue

            # Output filename
            if filename is not None:
                plot_filename = (
                    "{fn.parent}/{fn.stem}{i:0{w1}d}_{j:0{w2}d}{fn.suffix}".format(
                        fn=Path(filename),
                        i=i,
                        j=j,
                        w1=len(str(n_components)),
                        w2=len(str(n_modes)),
                    )
                )
            else:
                plot_filename = None

            # Fit a GMM to get class labels
            threshold = gmm.fit_gaussian_mixture(
                c,
                standardize=standardize,
                p_value=p_value,
                one_component_percentile=one_component_percentile,
                n_sigma=n_sigma,
                sklearn_kwargs=sklearn_kwargs,
                show_plot=show,
                plot_filename=plot_filename,
                plot_kwargs=plot_kwargs,
                log_message=False,
            )

            # Calculate the percentile from the threshold
            percentiles[i, j] = stats.percentileofscore(c, threshold)

    return np.squeeze(percentiles)


def threshold(
    conn_map,
    percentile,
    subtract_mean=False,
    mean_weights=None,
    absolute_value=False,
    return_edges=False,
):
    """Return edges that exceed a threshold.

    Parameters
    ---------
    conn_map : np.ndarray
        Connectivity matrix to threshold. Shape must be (n_components, n_modes,
        n_channels, n_channels), (n_modes, n_channels, n_channels) or
        (n_channels, n_channels).
    percentile : float or np.ndarray
        Percentile to threshold with. Should be between 0 and 100.
        Shape must be (n_components, n_modes), (n_modes,) or a float.
    subtract_mean : bool, optional
        Should we subtract the mean over modes before thresholding? The
        thresholding is only done to identify edges, the values returned in
        :code:`conn_map` are not mean subtracted.
    mean_weights : np.ndarray, optional
        Weights when calculating the mean over modes.
    absolute_value : bool, optional
        Should we take the absolute value before thresholding? The thresholding
        is only done to identify edges, the values returned in :code:`conn_map`
        are not absolute values. If :code:`subtract_mean=True`, the mean is
        subtracted before the absolute value.
    return_edges : bool, optional
        Should we return a boolean array for whether edges are above the
        threshold?

    Returns
    -------
    conn_map : np.ndarray
        Connectivity matrix with connections below the threshold set to zero.
        Or a boolean array if :code:`return_edges=True`. Shape is the same as
        the original :code:`conn_map`.
    """
    # Validation
    conn_map = array_ops.validate(
        conn_map,
        correct_dimensionality=4,
        allow_dimensions=[2, 3],
        error_message="conn_map must be of shape "
        + "(n_components, n_modes, n_channels, n_channels), "
        + "(n_modes, n_channels, n_channels) or (n_channels, n_channels)",
    )

    # Number of components and modes
    n_components = conn_map.shape[0]
    n_modes = conn_map.shape[1]
    n_channels = conn_map.shape[2]

    # Validatation
    if isinstance(percentile, float) or isinstance(percentile, int):
        percentile = percentile * np.ones([n_components, n_modes])

    percentile = array_ops.validate(
        percentile,
        correct_dimensionality=2,
        allow_dimensions=[0, 1],
        error_message="percentile must be of shape "
        + "(n_components, n_modes), (n_modes,) or float",
    )

    # Copy the original connectivity map
    c = conn_map.copy()

    # Subtract the mean
    if n_modes == 1:
        subtract_mean = False
    if subtract_mean:
        c -= np.average(c, axis=1, weights=mean_weights, keepdims=True)

    # Take absolute value
    if absolute_value:
        c = abs(c)

    # Set diagonal to nan
    c[:, :, range(n_channels), range(n_channels)] = np.nan

    # Are the connectivity matrices symmetric?
    c_is_symmetric = array_ops.check_symmetry(c, precision=1e-6)
    m, n = np.triu_indices(n_channels, k=1)

    # Which edges are greater than the threshold?
    edges = np.zeros(
        [n_components, n_modes, n_channels, n_channels],
        dtype=bool,
    )
    for i in range(n_components):
        for j in range(n_modes):
            if c_is_symmetric[i, j]:
                # We have a symmetric connectivity matrix
                # Threshold the upper triangle and copy to the lower triangle
                edges[i, j, m, n] = c[i, j, m, n] > np.nanpercentile(
                    c[i, j, m, n], percentile[i, j]
                )
                edges[i, j, n, m] = edges[i, j, m, n]
            else:
                # We have a directed connectivity matrix
                # Threshold each entry independently
                edges[i, j] = c[i, j] > np.nanpercentile(c[i, j], percentile[i, j])

    if return_edges:
        return np.squeeze(edges)

    # Zero the connections that are below the threshold
    conn_map[~edges] = 0

    return np.squeeze(conn_map)


def separate_edges(conn_map):
    """Separate positive and negative edges in a connectivity map.

    Parameters
    ----------
    conn_map : np.ndarray
        Connectivity map. Any shape.

    Returns
    -------
    pos_conn_map : np.ndarray
        Connectivity map with positive edges. Shape is the same as
        :code:`conn_map`.
    neg_conn_map : np.ndarray
        Connectivity map with negative edges. Shape is the same as
        :code:`conn_map`.
    """
    pos_conn_map = conn_map.copy()
    neg_conn_map = conn_map.copy()
    pos_conn_map[pos_conn_map < 0] = 0
    neg_conn_map[neg_conn_map > 0] = 0
    return pos_conn_map, neg_conn_map


def spectral_reordering(corr_mat):
    """Spectral re-ordering for correlation matrices.

    Parameters
    ----------
    corr_mat : np.ndarray
        Correlation matrix. Shape must be (n_channels, n_channels).

    Returns
    -------
    reorder_corr_mat : np.ndarray
        Re-ordered correlation matrix. Shape is (n_channels, n_channels).
    order : np.ndarray
        New ordering. Shape is (n_channels,).
    """
    # Add one to make all entries postive
    C = corr_mat + 1

    # Compute Q
    Q = -C
    np.fill_diagonal(Q, 0)
    Q -= np.sum(Q, axis=0)

    # Compute t
    t = np.diag(1.0 / np.sqrt(np.sum(C, axis=0)))

    # Compute D
    D = np.dot(np.dot(t, Q), t)

    # Eigevalue decomposition
    D, W = np.linalg.eig(D)
    v = W[:, 1]

    # Scale v
    v = np.dot(t, v)

    # Find permutations
    order = np.argsort(v)

    # Reorder
    reorder_corr_mat = corr_mat[order, :][:, order]

    return reorder_corr_mat, order


def save(
    connectivity_map,
    parcellation_file,
    filename=None,
    component=None,
    threshold=0,
    plot_kwargs=None,
    axes=None,
    combined=False,
    titles=None,
    n_rows=1,
):
    """Save connectivity maps as image files.

    This function is a wrapper for `nilearn.plotting.plot_connectome \
    <https://nilearn.github.io/stable/modules/generated/nilearn\
    .plotting.plot_connectome.html>`_.

    Parameters
    ----------
    connectivity_map : np.ndarray
        Matrices containing connectivity strengths to plot.
        Shape must be (n_components, n_modes, n_channels, n_channels),
        (n_modes, n_channels, n_channels) or (n_channels, n_channels).
    parcellation_file : str
        Name of parcellation file used.
    filename : str, optional
        Output filename. If :code:`None` is passed then the image is
        shown on screen. Must have extension :code:`.png`, :code:`.pdf`
        or :code:`.svg`.
    component : int, optional
        Spectral component to save.
    threshold : float or np.ndarray, optional
        Threshold to determine which connectivity to show. Should be between 0
        and 1. If a :code:`float` is passed the same threshold is used for all
        modes. Otherwise, threshold should be a numpy array of shape
        (n_modes,).
    plot_kwargs : dict, optional
        Keyword arguments to pass to the nilearn plotting function.
    axes : list, optional
        List of matplotlib axes to plot the connectivity maps on.
    combined : bool, optional
        Should the connectivity maps be combined on the same figure?
        The combined image is always shown on screen (for Juptyer notebooks).
        Note if :code:`True` is passed, the individual images will be deleted.
    titles : list, optional
        List of titles for each connectivity map. Only used if
        :code:`combined=True`.
    n_rows : int, optional
        Number of rows in the combined image. Only used if :code:`combined=True`.

    Examples
    --------
    Change colormap and views::

        connectivity.save(
            ...,
            plot_kwargs={
                "edge_cmap": "red_transparent_full_alpha_range",
                "display_mode": "lyrz",
            },
        )
    """
    # Suppress INFO messages from nibabel
    logging.getLogger("nibabel.global").setLevel(logging.ERROR)

    # Validation
    connectivity_map = np.copy(connectivity_map)
    error_message = (
        "Dimensionality of connectivity_map must be 3 or 4, "
        + f"got ndim={connectivity_map.ndim}."
    )
    connectivity_map = array_ops.validate(
        connectivity_map,
        correct_dimensionality=4,
        allow_dimensions=[2, 3],
        error_message=error_message,
    )

    if isinstance(threshold, float) or isinstance(threshold, int):
        threshold = np.array([threshold] * connectivity_map.shape[1])

    if np.any(threshold > 1) or np.any(threshold < 0):
        raise ValueError("threshold must be between 0 and 1.")

    if component is None:
        component = 0

    # Load parcellation file
    parcellation = Parcellation(parcellation_file)

    # Select the component we're plotting
    conn_map = connectivity_map[component]

    # Fill diagonal with zeros to help with the colorbar limits
    for c in conn_map:
        np.fill_diagonal(c, 0)

    # Default plotting settings
    default_plot_kwargs = {"node_size": 10, "node_color": "black"}

    # Loop through each connectivity map
    n_modes = conn_map.shape[0]
    axes = axes or [None] * n_modes
    output_files = []
    for i in trange(n_modes, desc="Saving images"):
        # Overwrite keyword arguments if passed
        kwargs = override_dict_defaults(default_plot_kwargs, plot_kwargs)

        # Output filename
        if filename is None:
            output_file = None
        else:
            output_file = "{fn.parent}/{fn.stem}{i:0{w}d}{fn.suffix}".format(
                fn=Path(filename), i=i, w=len(str(n_modes))
            )

        # If all connections are zero don't add a colourbar
        kwargs["colorbar"] = np.any(
            conn_map[i][~np.eye(conn_map[i].shape[-1], dtype=bool)] != 0
        )

        # Plot maps
        plotting.plot_connectome(
            conn_map[i],
            parcellation.roi_centers(),
            edge_threshold=f"{threshold[i] * 100}%",
            output_file=output_file,
            axes=axes[i],
            **kwargs,
        )
        output_files.append(output_file)

    if combined:
        # Combine the images
        if filename is None:
            raise ValueError("filename must be passed to save the combined image.")

        n_columns = -(n_modes // -n_rows)
        titles = titles or [None] * n_modes
        fig, axes = plt.subplots(n_rows, n_columns, figsize=(n_columns * 5, n_rows * 5))
        for i, ax in enumerate(axes.flatten()):
            ax.axis("off")
            if i < n_modes:
                ax.imshow(plt.imread(output_files[i]))
                ax.set_title(titles[i], fontsize=20)
        fig.tight_layout()
        fig.savefig(filename)

        # Remove the individual images
        for output_file in output_files:
            os.remove(output_file)


def save_interactive(
    connectivity_map,
    parcellation_file,
    filename=None,
    component=None,
    threshold=0,
    plot_kwargs=None,
):
    """Save connectivity maps as interactive HTML plots.

    This function is a wrapper for `nilearn.plotting.view_connectome \
    <https://nilearn.github.io/stable/modules/generated/nilearn\
    .plotting.view_connectome.html>`_

    Parameters
    ----------
    connectivity_map : np.ndarray
        Matrices containing connectivity strengths to plot.
        Shape must be (n_components, n_modes, n_channels, n_channels),
        (n_modes, n_channels, n_channels) or (n_channels, n_channels).
    parcellation_file : str
        Name of parcellation file used.
    filename : str, optional
        Output filename. If :code:`None` is passed then the image is
        shown on screen. Must have extension :code:`.html`.
    component : int, optional
        Spectral component to save.
    threshold : float or np.ndarray, optional
        Threshold to determine which connectivity to show. Should be between 0
        and 1. If a :code:`float` is passed the same threshold is used for all
        modes. Otherwise, threshold should be a numpy array of shape
        (n_modes,).
    plot_kwargs : dict, optional
        Keyword arguments to pass to the nilearn plotting function.
    """

    # Validation
    connectivity_map = np.copy(connectivity_map)
    error_message = (
        "Dimensionality of connectivity_map must be 3 or 4, "
        + f"got ndim={connectivity_map.ndim}."
    )
    connectivity_map = array_ops.validate(
        connectivity_map,
        correct_dimensionality=4,
        allow_dimensions=[2, 3],
        error_message=error_message,
    )

    if isinstance(threshold, float) or isinstance(threshold, int):
        threshold = np.array([threshold] * connectivity_map.shape[1])

    if np.any(threshold > 1) or np.any(threshold < 0):
        raise ValueError("threshold must be between 0 and 1.")

    if component is None:
        component = 0

    # Load parcellation file
    parcellation = Parcellation(parcellation_file)

    # Select the component we're plotting
    conn_map = connectivity_map[component]

    # Fill diagonal with zeros to help with the colorbar limits
    for c in conn_map:
        np.fill_diagonal(c, 0)

    # Default plotting settings
    default_plot_kwargs = {"node_size": 10, "node_color": "black"}

    # Loop through each connectivity map
    n_modes = conn_map.shape[0]
    for i in trange(n_modes, desc="Saving images"):
        # Overwrite keyword arguments if passed
        kwargs = override_dict_defaults(default_plot_kwargs, plot_kwargs)

        # Output filename
        if filename is None:
            output_file = None
        else:
            output_file = "{fn.parent}/{fn.stem}{i:0{w}d}{fn.suffix}".format(
                fn=Path(filename), i=i, w=len(str(n_modes))
            )

        # The colour bar range is determined by the max value in the matrix
        # we zero the diagonal so it's not included
        np.fill_diagonal(conn_map[i], val=0)

        # Plot thick lines for the connections
        if "linewidth" not in kwargs:
            kwargs["linewidth"] = 12

        # Plot maps
        connectome = plotting.view_connectome(
            conn_map[i],
            parcellation.roi_centers(),
            edge_threshold=f"{threshold[i] * 100}%",
            **kwargs,
        )
        if filename is not None:
            connectome.save_as_html(output_file)
        else:
            return connectome



"""Utility/helper classes and functions."""

from osl_dynamics.utils import misc
from osl_dynamics.utils import model
from osl_dynamics.utils import parcellation
from osl_dynamics.utils import plotting
from osl_dynamics.utils import topoplots

from osl_dynamics.utils.misc import set_random_seed



"""Functions to calculate and save network power maps."""

import os
import logging
from pathlib import Path

import numpy as np
import nibabel as nib
from tqdm.auto import trange
from pqdm.threads import pqdm
import matplotlib.pyplot as plt

from osl_dynamics import array_ops, files
from osl_dynamics.analysis.spectral import get_frequency_args_range
from osl_dynamics.utils import parcellation, plotting

_logger = logging.getLogger("osl-dynamics")


def sliding_window_power(
    data, window_length, step_size=None, power_type="mean", concatenate=False, n_jobs=1
):
    """Calculate sliding window power.

    Parameters
    ----------
    data : list or np.ndarray
        Time series data. Shape must be (n_sessions, n_samples, n_channels)
        or (n_samples, n_channels).
    window_length : int
        Window length in samples.
    step_size : int, optional
        Number of samples to slide the window along the time series.
        If :code:`None` is passed, then a 50% overlap is used.
    power_type : str, optional
        Type of power to calculate. Can be :code:`"mean"` or :code:`"var"`.
    concatenate : bool, optional
        Should we concatenate the sliding window power from each array
        into one big time series?
    n_jobs : int, optional
        Number of jobs to run in parallel. Default is 1.

    Returns
    -------
    sliding_window_power : list or np.ndarray
        Time series of power vectors. Shape is (n_sessions, n_windows,
        n_channels) or (n_windows, n_channels).
    """
    # Validation
    if power_type not in ["mean", "var"]:
        raise ValueError(f"power_type must be 'mean' or 'var', not {power_type}")

    if power_type == "var":
        metric = np.var
    else:
        metric = np.mean

    if step_size is None:
        step_size = window_length // 2

    if isinstance(data, np.ndarray):
        if data.ndim != 3:
            data = [data]

    # Helper function to calculate power
    def _swp(x):
        n_samples = x.shape[0]
        n_channels = x.shape[1]
        n_windows = (n_samples - window_length - 1) // step_size + 1

        # Preallocate an array fo hold moving average values
        swp = np.empty([n_windows, n_channels], dtype=np.float32)
        for i in range(n_windows):
            window_ts = x[i * step_size : i * step_size + window_length]
            swp[i] = metric(window_ts, axis=0)

        return swp

    # Setup keyword arguments to pass to the helper function
    kwargs = [{"x": x} for x in data]

    if len(data) == 1:
        _logger.info("Sliding window power")
        results = [_swp(**kwargs[0])]

    elif n_jobs == 1:
        results = []
        for i in trange(len(data), desc="Sliding window power"):
            results.append(_swp(**kwargs[i]))

    else:
        _logger.info(f"Sliding window power")
        results = pqdm(
            kwargs,
            _swp,
            argument_type="kwargs",
            n_jobs=n_jobs,
        )

    if concatenate or len(results) == 1:
        results = np.concatenate(results)

    return results


def variance_from_spectra(
    frequencies,
    power_spectra,
    components=None,
    frequency_range=None,
    method="mean",
):
    """Calculates variance from power spectra.

    Parameters
    ----------
    frequencies : np.ndarray
        Frequency axis of the PSDs. Only used if :code:`frequency_range` is
        given. Shape must be (n_freq,).
    power_spectra : np.ndarray
        Power/cross spectra for each channel.
        Shape must be (n_channels, n_channels) or (n_channels,).
    components : np.ndarray, optional
        Spectral components. Shape must be (n_components, n_freq).
    frequency_range : list, optional
        Frequency range in Hz to integrate the PSD over. Default is full range.
    method : str
        Should take the sum of the PSD over the frequency range
        (:code:`method="sum"`), the integral of the PSD (:code:`"integral"`),
        or take the average value of the PSD (:code:`method="mean"`).

    Returns
    -------
    var : np.ndarray
        Variance over a frequency band for each component of each mode.
        Shape is (n_components, n_modes, n_channels) or (n_modes, n_channels)
        or (n_channels,).
    """

    # Validation
    if power_spectra.ndim == 2:
        # PSDs were passed
        power_spectra = power_spectra[np.newaxis, np.newaxis, ...]
        n_sessions, n_modes, n_channels, n_freq = power_spectra.shape

    elif power_spectra.shape[-2] != power_spectra.shape[-3]:
        # PSDs were passed, check dimensionality
        error_message = (
            "A (n_channels, n_freq), (n_modes, n_channels, n_freq) or "
            + "(n_sessions, n_modes, n_channels, n_freq) array must be passed."
        )
        power_spectra = array_ops.validate(
            power_spectra,
            correct_dimensionality=4,
            allow_dimensions=[2, 3],
            error_message=error_message,
        )
        n_sessions, n_modes, n_channels, n_freq = power_spectra.shape

    else:
        # Cross spectra were passed, check dimensionality
        error_message = (
            "A (n_channels, n_channels, n_freq), "
            + "(n_modes, n_channels, n_channels, n_freq) or "
            + "(n_sessions, n_modes, n_channels, n_channels, n_freq) "
            + "array must be passed."
        )
        power_spectra = array_ops.validate(
            power_spectra,
            correct_dimensionality=5,
            allow_dimensions=[3, 4],
            error_message=error_message,
        )
        n_sessions, n_modes, n_channels, n_channels, n_freq = power_spectra.shape

    if components is not None and frequency_range is not None:
        raise ValueError(
            "Only one of the arguments components or frequency range can be passed."
        )

    if frequency_range is not None and frequencies is None:
        raise ValueError(
            "If frequency_range is passed, frequenices must also be passed."
        )

    if method not in ["mean", "sum", "integral"]:
        raise ValueError("method should be 'mean', 'sum' or 'integral'.")

    # Number of spectral components
    if components is None:
        n_components = 1
    else:
        n_components = components.shape[0]

    # Calculate power maps for each array
    var = []
    for i in range(n_sessions):
        # Get PSDs
        if power_spectra.shape[-2] == power_spectra.shape[-3]:
            # Cross-spectra densities were passed
            psd = power_spectra[i, :, range(n_channels), range(n_channels)]
            psd = np.swapaxes(psd, 0, 1)
        else:
            # Only the PSDs were passed
            psd = power_spectra[i]

        # Concatenate over modes
        psd = psd.reshape(-1, n_freq)
        psd = psd.real

        if components is not None:
            # Calculate PSD for each spectral component
            p = components @ psd.T
            for j in range(n_components):
                p[j] /= np.sum(components[j])

        else:
            # Integrate over the given frequency range
            if frequency_range is None:
                if method == "sum":
                    p = np.sum(psd, axis=-1)
                elif method == "integral":
                    df = frequencies[1] - frequencies[0]
                    p = np.sum(psd * df, axis=-1)
                else:
                    p = np.mean(psd, axis=-1)
            else:
                [min_arg, max_arg] = get_frequency_args_range(
                    frequencies, frequency_range
                )
                if method == "sum":
                    p = np.sum(psd[..., min_arg:max_arg], axis=-1)
                elif method == "integral":
                    df = frequencies[1] - frequencies[0]
                    p = np.sum(psd[..., min_arg:max_arg] * df, axis=-1)
                else:
                    p = np.mean(psd[..., min_arg:max_arg], axis=-1)

        p = p.reshape(n_components, n_modes, n_channels)
        var.append(p)

    return np.squeeze(var)


def save(
    power_map,
    mask_file,
    parcellation_file,
    filename=None,
    component=0,
    subtract_mean=False,
    mean_weights=None,
    plot_kwargs=None,
    combined=False,
    titles=None,
    n_rows=1,
):
    """Saves power maps.

    Parameters
    ----------
    power_map : np.ndarray
        Power map to save. Can be of shape: (n_components, n_modes, n_channels),
        (n_modes, n_channels) or (n_channels,). A (..., n_channels, n_channels)
        array can also be passed. Warning: this function
        cannot be used if :code:`n_modes=n_channels`.
    mask_file : str
        Mask file used to preprocess the training data.
    parcellation_file : str
        Parcellation file used to parcellate the training data.
    filename : str, optional
        Output filename. If extension is :code:`.nii.gz` the power map is saved
        as a NIFTI file. Or if the extension is :code:`png/svg/pdf`, it is saved
        as images. If :code:`None` is passed then the image is shown on screen
        and the Matplotlib objects are returned.
    component : int, optional
        Spectral component to save.
    subtract_mean : bool, optional
        Should we subtract the mean power across modes?
    mean_weights: np.ndarray, optional
        Numpy array with weightings for each mode to use to calculate the mean.
        Default is equal weighting.
    plot_kwargs : dict, optional
        Keyword arguments to pass to `osl_dynamics.utils.plotting\
        .plot_brain_surface <https://osl-dynamics.readthedocs.io/en/latest/\
        autoapi/osl_dynamics/utils/plotting/index.html#osl_dynamics.utils\
        .plotting.plot_brain_surface>`_.
    combined : bool, optional
        Should the individual plots be combined into a single image?
        The combined image is always shown on screen (for Juptyer notebooks).
        Note if :code:`True` is passed, the individual images will be deleted.
    titles : list, optional
        List of titles for each power plot.
    n_rows : int, optional
        Number of rows in the combined image. Only used if :code:`combined=True`.

    Returns
    -------
    figures : list of plt.figure
        List of Matplotlib figure object. Only returned if
        :code:`filename=None`.
    axes : list of plt.axis
        List of Matplotlib axis object(s). Only returned if
        :code:`filename=None`.
    """
    if plot_kwargs is None:
        plot_kwargs = {}

    # Validation
    if filename is not None:
        allowed_extensions = [".nii", ".nii.gz", ".png", ".svg", ".pdf"]
        if not any([ext in filename for ext in allowed_extensions]):
            raise ValueError(
                "filename must have one of following extensions: "
                + f"{' '.join(allowed_extensions)}."
            )
    mask_file = files.check_exists(mask_file, files.mask.directory)
    parcellation_file = files.check_exists(
        parcellation_file, files.parcellation.directory
    )

    # Create a copy of the power map so we don't modify it
    power_map = np.copy(power_map)

    # Reshape the power maps
    power_map = np.squeeze(power_map)
    if power_map.ndim > 1:
        if power_map.shape[-1] == power_map.shape[-2]:
            # A n_channels by n_channels array has been passed,
            # extract the diagonal (the np.copy is needed because np.diagonal
            # returns a read-only array (this started in NumPy 1.9))
            power_map = np.copy(np.diagonal(power_map, axis1=-2, axis2=-1))
            if power_map.ndim == 1:
                power_map = power_map[np.newaxis, ...]
    else:
        power_map = power_map[np.newaxis, ...]

    power_map = array_ops.validate(
        power_map,
        correct_dimensionality=3,
        allow_dimensions=[2],
        error_message="power_map.shape is incorrect",
    )

    # Subtract weighted mean
    n_modes = power_map.shape[1]
    if n_modes == 1:
        subtract_mean = False
    if subtract_mean:
        power_map -= np.average(power_map, axis=1, weights=mean_weights)[
            :, np.newaxis, ...
        ]

    # Select the component to plot
    power_map = power_map[component]

    if titles is None:
        titles = [None] * n_modes
    else:
        if len(titles) != n_modes:
            raise ValueError(
                f"Number of titles ({len(titles)}) does not match the number "
                f"of power maps ({n_modes})."
            )

    # Just display the power map
    if filename is None:
        figures, axes = [], []
        for i in trange(n_modes, desc="Saving images"):
            fig, ax = plotting.plot_brain_surface(
                power_map[i],
                mask_file=mask_file,
                parcellation_file=parcellation_file,
                title=titles[i],
                **plot_kwargs,
            )
            figures.append(fig)
            axes.append(ax)
        return figures, axes

    else:
        if ".nii" in filename:
            # Convert parcel values to voxel values
            power_map = [
                parcellation.parcel_vector_to_voxel_grid(
                    mask_file, parcellation_file, p
                )
                for p in power_map
            ]
            power_map = np.moveaxis(power_map, 0, -1)

            # Save as nii file
            _logger.info(f"Saving {filename}")
            mask = nib.load(mask_file)
            nii = nib.Nifti1Image(power_map, mask.affine, mask.header)
            nib.save(nii, filename)

        else:
            # Save each map as an image
            output_files = []
            for i in trange(n_modes, desc="Saving images"):
                output_file = "{fn.parent}/{fn.stem}{i:0{w}d}{fn.suffix}".format(
                    fn=Path(filename), i=i, w=len(str(n_modes))
                )
                plotting.plot_brain_surface(
                    power_map[i],
                    mask_file=mask_file,
                    parcellation_file=parcellation_file,
                    title=titles[i],
                    filename=output_file,
                    **plot_kwargs,
                )
                output_files.append(output_file)

            if combined:
                n_columns = -(n_modes // -n_rows)

                # Combine images into a single image
                fig, axes = plt.subplots(
                    n_rows, n_columns, figsize=(n_columns * 5, n_rows * 5)
                )
                for i, ax in enumerate(axes.flatten()):
                    ax.axis("off")
                    if i < n_modes:
                        ax.imshow(plt.imread(output_files[i]))
                fig.tight_layout()
                fig.savefig(filename)

                # Remove the individual images
                for output_file in output_files:
                    os.remove(output_file)


def multi_save(
    group_power_map,
    session_power_map,
    mask_file,
    parcellation_file,
    filename=None,
    sessions=None,
    subtract_mean=False,
    mean_weights=None,
    plot_kwargs=None,
):
    """Saves group level and array level power maps.

    When training session-specific models we want to plot the group-level map
    and session-specific deviations. This function is a wrapper for `power.save
    <https://osl-dynamics.readthedocs.io/en/latest/autoapi/osl_dynamics\
    /analysis/power/index.html#osl_dynamics.analysis.power.save>`_, which helps
    us plot power maps for session-specific models.

    Parameters
    ----------
    group_power_map : np.ndarray
        Group level power map to save.
        Can be of shape: (n_modes, n_channels) or (n_channels,). A (...,
        n_channels, n_channels) can also be passed. Warning: this function
        cannot be used if :code:`n_modes` is equal to :code:`n_channels`.
    session_power_map : np.ndarray
        Session-level power maps to save.
        Can be of shape: (n_sessions, n_modes, n_channels), (n_modes,
        n_channels) or (n_channels,). A (..., n_channels, n_channels) array can
        also be passed. Warning: this function cannot be used if
        :code:`n_modes=n_channels`.
    mask_file : str
        Mask file used to preprocess the training data.
    parcellation_file : str
        Parcellation file used to parcellate the training data.
    filename : str, optional
        Output filename. If extension is :code:`.nii.gz` the power map is saved
        as a NIFTI file. Or if the extension is :code:`png/svg/pdf`, it is saved
        as images. If :code:`None` is passed then the image is shown on screen
        and the Matplotlib objects are returned.
    sessions : list, optional
        List of session indices to be plot power maps for.
    subtract_mean : bool, optional
        Should we subtract the mean power across modes?
    mean_weights: np.ndarray, optional
        Numpy array with weightings for each mode to use to calculate the mean.
        Default is equal weighting.
    plot_kwargs : dict, optional
        Keyword arguments to pass to `osl_dynamics.utils.plotting\
        .plot_brain_surface <https://osl-dynamics.readthedocs.io/en/latest/\
        autoapi/osl_dynamics/utils/plotting/index.html#osl_dynamics.utils\
        .plotting.plot_brain_surface>`_.
    """
    # Create a copy of the power maps so we don't modify them
    group_power_map = np.copy(group_power_map)
    session_power_map = np.copy(session_power_map)

    # Validation
    if group_power_map.ndim > 1:
        if group_power_map.shape[-1] == group_power_map.shape[-2]:
            # np.copy is needed because np.diagonal returns a read only array
            group_power_map = np.copy(np.diagonal(group_power_map, axis1=-2, axis2=-1))
    else:
        group_power_map = group_power_map[np.newaxis, ...]

    if session_power_map.ndim > 1:
        if session_power_map.shape[-1] == session_power_map.shape[-2]:
            # np.copy is needed because np.diagonal returns a read only array
            session_power_map = np.copy(
                np.diagonal(session_power_map, axis1=-2, axis2=-1)
            )
    else:
        session_power_map = session_power_map[np.newaxis, ...]

    group_power_map = array_ops.validate(
        group_power_map,
        correct_dimensionality=2,
        allow_dimensions=[1],
        error_message="group_power_map.shape is incorrect.",
    )
    session_power_map = array_ops.validate(
        session_power_map,
        correct_dimensionality=3,
        allow_dimensions=[1, 2],
        error_message="session_power_map.shape is incorrect",
    )

    if group_power_map.shape[0] != session_power_map.shape[1]:
        raise ValueError("group and array level power maps must have the same n_modes.")

    # Subtract mean
    n_modes = group_power_map.shape[0]
    if n_modes == 1:
        subtract_mean = False
    if subtract_mean:
        mean_group_power = np.average(
            group_power_map,
            axis=0,
            weights=mean_weights,
        )
        group_power_map -= mean_group_power[np.newaxis, ...]
        session_power_map -= mean_group_power[np.newaxis, np.newaxis, ...]

    # Save the group power map
    filename = Path(filename)
    group_dir = f"{filename.parent}/group"
    os.makedirs(group_dir, exist_ok=True)
    group_filename = f"{group_dir}/{filename.stem}{filename.suffix}"

    _logger.info("Saving group level power map:")
    save(
        group_power_map,
        filename=group_filename,
        mask_file=mask_file,
        parcellation_file=parcellation_file,
        plot_kwargs=plot_kwargs,
    )

    # Save the session-level power maps
    n_sessions = session_power_map.shape[0]
    if sessions is None:
        sessions = np.arange(n_sessions)

    for sess in sessions:
        session_dir = "{fn.parent}/sess_{sess:0{v}d}".format(
            fn=filename, sess=sess, v=len(str(n_sessions))
        )
        os.makedirs(session_dir, exist_ok=True)
        session_filename = f"{session_dir}/{filename.stem}{filename.suffix}"

        _logger.info(f"Saving session {sess} power map:")
        save(
            session_power_map[sess],
            filename=session_filename,
            mask_file=mask_file,
            parcellation_file=parcellation_file,
            plot_kwargs=plot_kwargs,
        )


def independent_components_to_volumetric_maps(
    ica_spatial_maps,
    ic_values,
    output_file=None,
):
    """Independent components to volumetric spatial map.

    Project the ic_values map to a brain map according to spatial
    maps of group ICA components. For example, ic_values can be the mean
    activation of states, or the rank-one decomposition of FC matrix
    corresponding to different states.

    Parameters
    ----------
    ica_spatial_maps : str or Nifti1Image
        Path to nifti file or nifti object containing the spatial maps
        obtained from group ICA. Should be in (n_x, n_y, n_z, n_voxels)
        format.
    ic_values : np.ndarray
        Independent component values.
        Should be in (n_maps, n_ica_components) format.
    output_file : str, optional
        Path to output file to save volumetric map to.

    Returns
    -------
    vol_map : Nifti1Image
        Volumetric maps. Shape is (n_x, n_y, n_z, n_maps).
    """
    # Load nifti
    if isinstance(ica_spatial_maps, str):
        ica_spatial_maps = nib.load(ica_spatial_maps)

    # Validation
    if not isinstance(ica_spatial_maps, nib.nifti1.Nifti1Image):
        raise ValueError(
            "ica_spatial_maps must be a path to a nifti file or "
            "nib.nifti1.Nifti1Image object."
        )

    # Get data from nifti
    ica_spatial_maps_data = ica_spatial_maps.get_fdata()
    ica_spatial_maps_shape = ica_spatial_maps_data.shape
    n_maps, n_ica_components = ic_values.shape

    # Calculate volumetric maps
    data = np.matmul(ica_spatial_maps_data, ic_values.T)

    # Store map to a Nifti1Image type
    vol_map = nib.nifti1.Nifti1Image(
        data,
        affine=ica_spatial_maps.affine,
        header=ica_spatial_maps.header,
    )

    if output_file is not None:
        # Save
        _logger.info(f"Saving {output_file}")
        nib.save(vol_map, output_file)

    return vol_map


def independent_components_to_surface_maps(
    ica_spatial_maps, ic_values, output_file=None
):
    """Independent components to surface spatial map.

    Project the ic_values map to a surface map according to spatial maps
    of group ICA components. For example, ic_values can be the mean
    activation of states, or the rank-one decomposition of FC matrix
    corresponding to different states.

    Different from independent_components_to_volumetric_maps,
    this function works on CIFTI spatial maps.

    Parameters
    ----------
    ica_spatial_maps : str or Cifti2Image
        Path to cifti file or cifti object containing the spatial maps
        obtained from group ICA. Should be in (n_ica_components, n_voxels)
        format.
    ic_values: np.ndarray
        Independent component values.
        Should be in (n_maps, n_ica_components) format.
    output_file : str, optional
        Path to output file to save surface map to.

    Returns
    -------
    surf_map : Cifti2Image
        Surface map. Shape is (n_maps, n_voxels).
    """
    # Load cifti
    if isinstance(ica_spatial_maps, str):
        ica_spatial_maps = nib.load(ica_spatial_maps)

    # Validation
    if not isinstance(ica_spatial_maps, nib.cifti2.cifti2.Cifti2Image):
        raise ValueError(
            "ica_spatial_maps must be a path to a cifti file or "
            "nib.cifti2.cifti2.Cifti2Image object."
        )

    # Get data from cifti
    ica_spatial_maps_data = ica_spatial_maps.get_fdata()
    header = ica_spatial_maps.header
    n_maps = ic_values.shape[0]

    # Calculate surface maps
    data = np.matmul(ic_values, ica_spatial_maps_data)

    # Define the new axes: the Map i (i = 1 ,..., n_maps)
    axes = nib.cifti2.cifti2_axes.ScalarAxis([f"Map {i + 1}" for i in range(n_maps)])

    # Define the new header using axes + axes[1] from old header
    header = nib.cifti2.cifti2.Cifti2Header.from_axes((axes, header.get_axis(1)))

    # Combine new data and new header to obtain new CIFTI object
    surf_map = nib.cifti2.cifti2.Cifti2Image(data, header)

    if output_file is not None:
        # Save
        _logger.info(f"Saving {output_file}")
        nib.save(surf_map, output_file)

    return surf_map



"""Miscellaneous utility classes and functions."""

import inspect
import logging
import pickle
import random
import sys
from copy import copy
from pathlib import Path
from contextlib import contextmanager

import numpy as np
import yaml
from yaml.constructor import ConstructorError

_logger = logging.getLogger("osl-dynamics")


def nextpow2(x):
    """Next power of 2.

    Parameters
    ----------
    x : int
        Any integer.

    Returns
    -------
    res : int
        The smallest power of two that is greater than or equal to the absolute
        value of x.
    """
    if x == 0:
        return 0
    res = np.ceil(np.log2(np.abs(x)))
    return res.astype("int")


def leading_zeros(number, largest_number):
    """Pad a number with leading zeros.

    This is useful for creating a consistent naming scheme for files.

    Parameters
    ----------
    number : int
        Number to be padded.
    largest_number : int
        Largest number in the set.

    Returns
    -------
    padded_number : str
        Number padded with leading zeros.
    """
    min_length = len(str(largest_number))
    padded_number = str(number).zfill(min_length)
    return padded_number


def override_dict_defaults(default_dict, override_dict=None):
    """Helper function to update default dictionary values with user values.

    Parameters
    ----------
    default_dict : dict
        Dictionary of default values.
    override_dict : dict, optional
        Dictionary of user values.

    Returns
    -------
    new_dict : dict
        default_dict with values replaced by user values.
    """
    if override_dict is None:
        override_dict = {}
    return {**default_dict, **override_dict}


def listify(obj):
    """Create a list from any input.

    If :code:`None` is passed, return an empty list.
    If a list is passed, return the list.
    If a tuple is passed, return it as a list.
    If any other object is passed, return it as a single item list.

    Parameters
    ----------
    obj : typing.Any
        Object to be transformed to a list.

    Returns
    -------
    Object as a list.
    """
    if obj is None:
        return []
    if isinstance(obj, list):
        return obj
    if isinstance(obj, tuple):
        return list(obj)
    else:
        return [obj]


def replace_argument(func, name, item, args, kwargs, append=False):
    """Replace arguments in function calls.

    Parameters
    ----------
    func : callable
        The function being called.
    name : str
        Name of the variable to be modified.
    item
        The value to be added.
    args : dict
        Original arguments.
    kwargs : dict
        Original keyword arguments.
    append : bool, optional
        Whether the value should be appended or replace the existing argument.

    Returns
    -------
    args : list
        Arguments.
    kwargs : dict
        Keyword arguments.
    """
    args = copy(listify(args))
    kwargs = copy(kwargs)
    param_order = list(inspect.signature(func).parameters)
    param_position = param_order.index(name)
    if len(args) > param_position:
        if append:
            args[param_position] = listify(args[param_position]) + listify(item)
        else:
            args[param_position] = item
    elif name in kwargs:
        if append:
            kwargs[name] = listify(kwargs[name]) + listify(item)
        else:
            kwargs[name] = item
    else:
        kwargs[name] = item
    return args, kwargs


def get_argument(func, name, args, kwargs):
    """Get an argument passed to a function call whether it is a normal
    argument or keyword argument.

    Parameters
    ----------
    func : callable
        The function being called.
    name : str
        Name of the variable to be modified.
    args : dict
        Arguments.
    kwargs : dict
        Keyword arguments.

    Returns
    -------
    args : argument
        Argument.
    """
    args = copy(listify(args))
    kwargs = copy(kwargs)
    param_order = list(inspect.signature(func).parameters)
    param_position = param_order.index(name)
    if len(args) > param_position:
        arg = args[param_position]
    else:
        if name not in kwargs:
            return None
        arg = kwargs[name]
    return arg


def check_arguments(args, kwargs, index, name, value, comparison_op):
    """Checks the arguments passed to a function.

    Parameters
    ----------
    args : list
        Arguments.
    kwargs : dict
        Keyword arguments.
    index : int
        Index of argument.
    name : str
        Name of keyword argument.
    value
        Value to compare to given argument.
    comparison_op : func
        Comparison operation for checking the original.

    Returns
    -------
    valid : bool
        If the given value is valid as determined by the comparison operation.
    """

    # Check if the argument we want to check is a normal argument
    args = listify(args)
    if len(args) >= index:
        return comparison_op(args[index], value)

    # Check if it is a keyword argument
    elif name in kwargs:
        return comparison_op(kwargs[name], value)

    # Otherwise the argument we want to check isn't in args or kwargs
    else:
        return False


def array_to_memmap(filename, array):
    """Save an array and reopen it as a np.memmap.

    Parameters
    ----------
    filename : str
        The name of the file to save to.
    array : np.ndarray
        The array to save.

    Returns
    -------
    memmap : np.memmap
        Memory map.
    """
    path = Path(filename)
    if path.exists():
        # Delete npy file
        path.unlink()

    # Save array
    np.save(filename, array)

    # Load as a memmap
    return np.load(filename, mmap_mode="r+")


class MockFlags:
    """Flags for memmap header construction.

    Parameters
    ----------
    shape : list of int
        The shape of the array being mapped.
    c_contiguous : bool, optional
        Is the array C contiguous or F contiguous?
    """

    def __init__(self, shape, c_contiguous=True):
        self.c_contiguous = c_contiguous
        self.f_contiguous = (not c_contiguous) or (c_contiguous and len(shape) == 1)


class MockArray:
    """Create an empty array on disk without creating it in memory.

    Parameters
    ----------
    shape : list of int
        Dimensions or the array being created.
    dtype : type
        The data type of the array.
    c_contiguous : bool, optional
        Is the array C contiguous or F contiguous?
    """

    def __init__(self, shape, dtype=np.float64, c_contiguous=True):
        self.shape = shape
        self.dtype = np.dtype(dtype)
        self.flags = MockFlags(shape, c_contiguous)

        self.filename = None

    def save(self, filename):
        if filename[-4:] != ".npy":
            filename = f"{filename}.npy"
        self.filename = filename
        if self.dtype.itemsize == 0:
            buffer_size = 0
        else:
            # Set buffer size to 16 MiB to hide the Python loop overhead.
            buffer_size = max(16 * 1024**2 // self.dtype.itemsize, 1)

        n_chunks, remainder = np.divmod(
            np.product(self.shape) * self.dtype.itemsize, buffer_size
        )

        with open(filename, "wb") as f:
            np.lib.format.write_array_header_2_0(
                f, np.lib.format.header_data_from_array_1_0(self)
            )

            for chunk in range(n_chunks):
                f.write(b"\x00" * buffer_size)
            f.write(b"\x00" * remainder)

    def memmap(self):
        if self.filename is None:
            raise ValueError("filename has not been provided.")
        return np.load(self.filename, mmap_mode="r+")

    @classmethod
    def to_disk(cls, filename, shape, dtype=np.float64, c_contiguous=True):
        mock_array = cls(shape, dtype, c_contiguous)
        mock_array.save(filename)

    @classmethod
    def get_memmap(cls, filename, shape, dtype=np.float64, c_contiguous=True):
        cls.to_disk(filename, shape, dtype, c_contiguous)
        return np.load(filename, mmap_mode="r+")


class NumpyLoader(yaml.UnsafeLoader):
    def find_python_name(self, name, mark, unsafe=False):
        if not name:
            raise ConstructorError(
                "while constructing a Python object",
                mark,
                "expected non-empty name appended to the tag",
                mark,
            )
        if "." in name:
            module_name, object_name = name.rsplit(".", 1)
        else:
            module_name = "builtins"
            object_name = name
        if "numpy" in module_name:
            try:
                __import__(module_name)
            except ImportError as exc:
                raise ConstructorError(
                    "while constructing a Python object",
                    mark,
                    "cannot find module %r (%s)" % (module_name, exc),
                    mark,
                )
        if module_name not in sys.modules:
            raise ConstructorError(
                "while constructing a Python object",
                mark,
                "module %r is not imported" % module_name,
                mark,
            )
        module = sys.modules[module_name]
        if not hasattr(module, object_name):
            raise ConstructorError(
                "while constructing a Python object",
                mark,
                "cannot find %r in the module %r" % (object_name, module.__name__),
                mark,
            )
        return getattr(module, object_name)


def save(filename, array):
    """Save a file.

    Parameters
    ----------
    filename : str
        Path to file to save to. Must be '.npy' or '.pkl'.
    array : np.ndarray or list
        Array to save.
    """
    # Validation
    ext = Path(filename).suffix
    if ext not in [".npy", ".pkl"]:
        raise ValueError("filename extension must be .npy or .pkl.")

    # Save
    _logger.info(f"Saving {filename}")
    if ext == ".pkl":
        pickle.dump(array, open(filename, "wb"))
    else:
        np.save(filename, array)


def load(filename, **kwargs):
    """Load a file.

    Parameters
    ----------
    filename : str
        Path to file to load. Must be '.npy' or '.pkl'.

    Returns
    -------
    array : np.ndarray or list
        Array loaded from the file.
    """
    # Validation
    ext = Path(filename).suffix
    if ext not in [".npy", ".pkl"]:
        raise ValueError("filename extension must be .npy or .pkl.")

    # Load
    _logger.info(f"Loading {filename}")
    if ext == ".pkl":
        array = pickle.load(open(filename, "rb"))
    else:
        array = np.load(filename, **kwargs)

    return array


def set_random_seed(seed, op_determinism=False):
    """Set all random seeds.

    This includes Python's random module, NumPy and TensorFlow.

    Parameters
    ----------
    seed : int
        Random seed.
    op_determinism : bool, optional
        Whether to enable operation determinism in TensorFlow.
        If True, TensorFlow operations will be deterministic, generally at the cost of
        lower performance. Note that the model may run slower if enabled.
    """
    import tensorflow as tf  # avoids slow imports

    _logger.info(f"Setting random seed to {seed}")

    tf.keras.utils.set_random_seed(seed)
    if op_determinism:
        tf.config.experimental.enable_op_determinism()


@contextmanager
def set_logging_level(logger, level):
    current_level = logger.getEffectiveLevel()
    try:
        logger.setLevel(level)
        yield
    finally:
        logger.setLevel(current_level)
